"""
This file holds all of the CLI commands for the "anyscale machines" path.
"""
from typing import Optional

import click
from rich import print_json
import tabulate

from anyscale.controllers.machine_controller import MachineController


@click.group(
    "machine", help="Commands to interact with machines in Anyscale.",
)
def machine_cli() -> None:
    pass


@machine_cli.command(name="list", help="List machines registered to Anyscale.")
@click.option("--cloud", type=str, help="Provide a cloud name.")
@click.option("--cloud-id", type=str, help="Provide a cloud ID.")
@click.option(
    "--format",
    "format_",
    type=click.Choice(["table", "json"]),
    default="table",
    help="The format to return results in.",
)
def list_machines(cloud: Optional[str], cloud_id: Optional[str], format_: str) -> None:

    if not cloud and not cloud_id:
        raise ValueError("One of {'cloud', 'cloud-id'} must be provided.")

    machines_controller = MachineController()
    output = machines_controller.list_machines(cloud=cloud, cloud_id=cloud_id)

    if format_ == "json":
        rows = []
        for m in output.machines:
            rows.append(
                {
                    "machine_id": m.machine_id,
                    "hostname": m.hostname,
                    "machine_shape": m.machine_shape,
                    "connection_state": m.connection_state,
                    "allocation_state": m.allocation_state,
                    "cluster_id": m.cluster_id,
                    "cluster_creator_name": m.cluster_creator_name,
                    "cluster_name": m.cluster_name,
                    "anyscale_version": m.anyscale_version,
                }
            )
        print_json(data=rows)
    elif format_ == "table":
        table = []
        columns = [
            "MACHINE",
            "HOSTNAME",
            "SHAPE",
            "HEALTHY",
            "STATUS",
            "OWNER",
            "CLUSTER ID",
            "CLUSTER NAME",
            "ANYSCALE VERSION",
        ]
        for m in output.machines:
            table.append(
                [
                    m.machine_id,
                    m.hostname,
                    m.machine_shape,
                    "Yes" if m.connection_state == "connected" else "No",
                    m.allocation_state.title(),
                    m.cluster_creator_name,
                    m.cluster_id,
                    m.cluster_name,
                    m.anyscale_version,
                ]
            )

        print(
            tabulate.tabulate(table, tablefmt="plain", headers=columns, stralign="left")
        )


@machine_cli.command(name="delete", help="Delete a machine from a machine pool.")
@click.option("--cloud", type=str, help="Provide a cloud name.")
@click.option("--cloud-id", type=str, help="Provide a cloud ID.")
@click.option("--machine-id", type=str, help="Provide a machine ID.")
def delete_machine(
    cloud: Optional[str], cloud_id: Optional[str], machine_id: str
) -> None:

    if not cloud and not cloud_id:
        raise ValueError("One of {'cloud', 'cloud-id'} must be provided.")

    machines_controller = MachineController()
    machines_controller.delete_machine(
        cloud=cloud, cloud_id=cloud_id, machine_id=machine_id,
    )

    print(f"Machine {machine_id} has been deleted.")



from unittest.mock import Mock, patch

from click.testing import CliRunner
import pytest

from anyscale.client.openapi_client.models import ClusterManagementStackVersions
from anyscale.commands.cloud_commands import register_cloud, setup_cloud


@pytest.mark.parametrize("enable_head_node_fault_tolerance", [True, False])
@pytest.mark.parametrize("yes", [True, False])
def test_cloud_setup(enable_head_node_fault_tolerance, yes):
    runner = CliRunner()
    mock_cloud_controller = Mock()
    mock_setup_managed_cloud = Mock()
    mock_cloud_controller().setup_managed_cloud = mock_setup_managed_cloud

    provider = "aws"
    region = "us-west-2"
    name = "unit-test-cloud"

    args = ["--provider", provider, "--name", name, "--region", region]
    if yes:
        args.append("--yes")

    if enable_head_node_fault_tolerance:
        args.append("--enable-head-node-fault-tolerance")

    with patch(
        "anyscale.commands.cloud_commands.CloudController", new=mock_cloud_controller,
    ):
        runner.invoke(setup_cloud, args=args)

    mock_setup_managed_cloud.assert_called_once_with(
        provider=provider,
        region=region,
        name=name,
        functional_verify=None,
        cluster_management_stack_version=ClusterManagementStackVersions.V2,
        enable_head_node_fault_tolerance=enable_head_node_fault_tolerance,
        yes=yes,
    )


@pytest.mark.parametrize(
    (("enable_head_node_fault_tolerance", "correct_memorystore_name")),
    [pytest.param(True, False), pytest.param(True, True), pytest.param(False, True)],
)
@pytest.mark.parametrize("host_project_id", [None, "host-project-id"])
def test_cloud_register(
    enable_head_node_fault_tolerance, correct_memorystore_name, caplog, host_project_id
):
    runner = CliRunner()
    mock_cloud_controller = Mock()
    mock_register_cloud = Mock()
    mock_cloud_controller().register_gcp_cloud = mock_register_cloud

    provider = "gcp"
    region = "us-west2"
    name = "unit-test-cloud"
    project_id = "mock_project_id"
    vpc_name = "vpc-cld-advbe8prrhdxx7cgnrgimtp8f4"
    subnet_names = "subnet-cld-advbe8prrhdxx7cgnrgimtp8f4"
    firewall_policy_names = "firewall-policy-cld-advbe8prrhdxx7cgnrgimtp8f4"
    cloud_storage_bucket_name = "storage-bucket-cld-advbe8prrhdxx7cgnrgimtp8f4"
    filestore_instance_id = "filestore-cld-advbe8prrhdxx7cgnrgimtp8f4"
    filestore_location = "us-west1-b"
    anyscale_service_account_email = (
        "anyscale-access-b11959be@allen-dev-396002.iam.gserviceaccount.com"
    )
    instance_service_account_email = (
        "cld-advbe8prrhdxx7cgnrgimtp8f4@allen-dev-396002.iam.gserviceaccount.com"
    )
    provider_name = "projects/604561082599/locations/global/workloadIdentityPools/anyscale-provider-pool-b11959be/providers/anyscale-access"
    if correct_memorystore_name:
        memorystore_instance_name = "projects/allen-dev-396002/locations/us-west1/instances/redis-cld-advbe8prrhdxx7cgnrgimtp8"
    else:
        memorystore_instance_name = "redis-cld-advbe8prrhdxx7cgnrgimtp8"

    args = [
        "--provider",
        provider,
        "--name",
        name,
        "--region",
        region,
        "--project-id",
        project_id,
        "--vpc-name",
        vpc_name,
        "--subnet-names",
        subnet_names,
        "--firewall-policy-names",
        firewall_policy_names,
        "--cloud-storage-bucket-name",
        cloud_storage_bucket_name,
        "--filestore-instance-id",
        filestore_instance_id,
        "--filestore-location",
        filestore_location,
        "--anyscale-service-account-email",
        anyscale_service_account_email,
        "--instance-service-account-email",
        instance_service_account_email,
        "--provider-name",
        provider_name,
    ]
    if host_project_id:
        args.extend(["--host-project-id", host_project_id])

    if enable_head_node_fault_tolerance:
        args.extend(["--memorystore-instance-name", memorystore_instance_name])

    with patch(
        "anyscale.commands.cloud_commands.CloudController", new=mock_cloud_controller,
    ):
        if not correct_memorystore_name:
            result = runner.invoke(register_cloud, args=args, catch_exceptions=False)
            assert "Please provide a valid memorystore instance name." in result.output
            mock_register_cloud.assert_not_called()
        else:
            runner.invoke(register_cloud, args=args)

            mock_register_cloud.assert_called_once_with(
                region=region,
                name=name,
                project_id=project_id,
                vpc_name=vpc_name,
                subnet_names=subnet_names.split(","),
                filestore_instance_id=filestore_instance_id,
                filestore_location=filestore_location,
                anyscale_service_account_email=anyscale_service_account_email,
                instance_service_account_email=instance_service_account_email,
                provider_id=provider_name,
                firewall_policy_names=firewall_policy_names.split(","),
                cloud_storage_bucket_name=cloud_storage_bucket_name,
                memorystore_instance_name=memorystore_instance_name
                if enable_head_node_fault_tolerance
                else None,
                host_project_id=host_project_id,
                functional_verify=None,
                private_network=False,
                cluster_management_stack_version=ClusterManagementStackVersions.V2,
                yes=False,
                skip_verifications=False,
            )



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
#### Testing notes
# To avoid blowing up on the printing of a string value, place the following at the beginning of imports:
# import anyscale.utils.protected_string
# anyscale.utils.protected_string.FAIL_ON_MISUSE = False
#
# When testing routers, please pass in a JSON object with the string & let pydantic construct the ProtectedString
# See test_users_router.py::test_login_user for some examples.

import copy
import hashlib
from typing import Any, cast, Dict, List, Type, Union
import warnings

import argon2

from anyscale.shared_anyscale_utils.conf import ANYSCALE_ENV


hasher = argon2.PasswordHasher()
# Whether to hard fail on serializing
FAIL_ON_MISUSE = ANYSCALE_ENV != "production"


class FastApiMixIn:
    @classmethod
    def __get_validators__(cls) -> Any:
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema: Any) -> Any:
        # __modify_schema__ should mutate the dict it receives in place, the
        # returned value will be ignored
        field_schema.update({"type": "string"})

    @classmethod
    def validate(cls, v: Any) -> Any:
        if not isinstance(v, str):
            raise TypeError("string required")
        return cls(v)  # type: ignore


class ProtectedString(FastApiMixIn):
    def __init__(self, s: Union[str, "ProtectedString"]):
        if isinstance(s, ProtectedString):
            self._UNSAFE_DO_NOT_USE: str = s._UNSAFE_DO_NOT_USE  # noqa: SLF001
        elif isinstance(s, str):
            self._UNSAFE_DO_NOT_USE = s
        else:
            raise ValueError(
                f"{type(s)}:'{repr(s)}' is not a valid type for ProtectedString"
            )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ProtectedString):
            return self._UNSAFE_DO_NOT_USE == other._UNSAFE_DO_NOT_USE  # noqa: SLF001
        return False

    def __str__(self) -> str:
        # This will collide like crazy (~1% chance of collision in 3000 guesses), but since the goal here is human debuggability, that is probably fine.
        m = hashlib.blake2b(self._UNSAFE_DO_NOT_USE.encode("utf-8"), digest_size=4)

        error_msg = "ProtectedStrings should not be serialized directly."
        if FAIL_ON_MISUSE:
            raise RuntimeError(error_msg)
        else:
            warnings.warn(error_msg)
        return f"<{self.__class__.__name__} digest={m.hexdigest()}>"

    def __repr__(self) -> str:
        return str(self)


class UserPassword(ProtectedString):
    # Return ProtectedString
    def hash(self) -> "UserPasswordHash":
        return UserPasswordHash(hasher.hash(self._UNSAFE_DO_NOT_USE))

    # Takes a ProtectedString
    def verify_against_hash(self, other_hash: "UserPasswordHash") -> bool:
        return cast(
            bool,  # black: no
            hasher.verify(
                other_hash._UNSAFE_DO_NOT_USE, self._UNSAFE_DO_NOT_USE  # noqa: SLF001
            ),
        )


class Token(UserPassword):
    """See `UserPassword` implementation"""

    # NOTE: Currently this cannot be a field in Pydantic models.
    # The model creation will error b/c it expects it to be of type `str`


class UserPasswordHash(ProtectedString):
    def __init__(self, value: ProtectedString):
        super().__init__(value)

        # Only method we can find that will validate that the value is actually
        # an argon2 hash
        self.needs_rehash()

    def needs_rehash(self) -> bool:
        return cast(bool, hasher.check_needs_rehash(self._UNSAFE_DO_NOT_USE))


class HashedToken(UserPasswordHash):
    pass


def protect_strings(
    input_dictionary: Dict[str, Any],
    keys: Union[List[str], str],
    protected_string_type: Type[ProtectedString],  # Callable[[str], ProtectedString]
    allow_optional_strings: bool = False,
) -> Dict[str, Any]:
    result_dict = copy.deepcopy(input_dictionary)
    if isinstance(keys, str):
        keys = [keys]
    for key in keys:
        old_string = result_dict[key]
        if isinstance(old_string, str):
            result_dict[key] = protected_string_type(old_string)
        elif allow_optional_strings and old_string is None:
            result_dict[key] = None
        else:
            raise ValueError("Tried to protect a non-string")

    return result_dict



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class RayRuntimeEnvConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'working_dir': 'str',
        'py_modules': 'list[str]',
        'pip': 'list[str]',
        'conda': 'object',
        'env_vars': 'dict(str, str)'
    }

    attribute_map = {
        'working_dir': 'working_dir',
        'py_modules': 'py_modules',
        'pip': 'pip',
        'conda': 'conda',
        'env_vars': 'env_vars'
    }

    def __init__(self, working_dir=None, py_modules=None, pip=None, conda=None, env_vars=None, local_vars_configuration=None):  # noqa: E501
        """RayRuntimeEnvConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._working_dir = None
        self._py_modules = None
        self._pip = None
        self._conda = None
        self._env_vars = None
        self.discriminator = None

        if working_dir is not None:
            self.working_dir = working_dir
        if py_modules is not None:
            self.py_modules = py_modules
        if pip is not None:
            self.pip = pip
        if conda is not None:
            self.conda = conda
        if env_vars is not None:
            self.env_vars = env_vars

    @property
    def working_dir(self):
        """Gets the working_dir of this RayRuntimeEnvConfig.  # noqa: E501

        The working directory that your code will run in. Must be a remote URI like an s3 or git path.  # noqa: E501

        :return: The working_dir of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: str
        """
        return self._working_dir

    @working_dir.setter
    def working_dir(self, working_dir):
        """Sets the working_dir of this RayRuntimeEnvConfig.

        The working directory that your code will run in. Must be a remote URI like an s3 or git path.  # noqa: E501

        :param working_dir: The working_dir of this RayRuntimeEnvConfig.  # noqa: E501
        :type: str
        """

        self._working_dir = working_dir

    @property
    def py_modules(self):
        """Gets the py_modules of this RayRuntimeEnvConfig.  # noqa: E501

        Python modules that will be installed along with your runtime env. These must be remote URIs.  # noqa: E501

        :return: The py_modules of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._py_modules

    @py_modules.setter
    def py_modules(self, py_modules):
        """Sets the py_modules of this RayRuntimeEnvConfig.

        Python modules that will be installed along with your runtime env. These must be remote URIs.  # noqa: E501

        :param py_modules: The py_modules of this RayRuntimeEnvConfig.  # noqa: E501
        :type: list[str]
        """

        self._py_modules = py_modules

    @property
    def pip(self):
        """Gets the pip of this RayRuntimeEnvConfig.  # noqa: E501

        A list of pip packages to install.  # noqa: E501

        :return: The pip of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._pip

    @pip.setter
    def pip(self, pip):
        """Sets the pip of this RayRuntimeEnvConfig.

        A list of pip packages to install.  # noqa: E501

        :param pip: The pip of this RayRuntimeEnvConfig.  # noqa: E501
        :type: list[str]
        """

        self._pip = pip

    @property
    def conda(self):
        """Gets the conda of this RayRuntimeEnvConfig.  # noqa: E501

        [Union[Dict[str, Any], str]: Either the conda YAML config or the name of a local conda env (e.g., \"pytorch_p36\"),   # noqa: E501

        :return: The conda of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: object
        """
        return self._conda

    @conda.setter
    def conda(self, conda):
        """Sets the conda of this RayRuntimeEnvConfig.

        [Union[Dict[str, Any], str]: Either the conda YAML config or the name of a local conda env (e.g., \"pytorch_p36\"),   # noqa: E501

        :param conda: The conda of this RayRuntimeEnvConfig.  # noqa: E501
        :type: object
        """

        self._conda = conda

    @property
    def env_vars(self):
        """Gets the env_vars of this RayRuntimeEnvConfig.  # noqa: E501

        Environment variables to set.  # noqa: E501

        :return: The env_vars of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this RayRuntimeEnvConfig.

        Environment variables to set.  # noqa: E501

        :param env_vars: The env_vars of this RayRuntimeEnvConfig.  # noqa: E501
        :type: dict(str, str)
        """

        self._env_vars = env_vars

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, RayRuntimeEnvConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, RayRuntimeEnvConfig):
            return True

        return self.to_dict() != other.to_dict()



from contextlib import nullcontext
from datetime import datetime
import tempfile
from typing import Any, Callable, Optional
from unittest.mock import Mock, patch

from click import ClickException
import pytest
import yaml

from anyscale.cluster_env import get_build_from_cluster_env_identifier
from anyscale.controllers.cluster_env_controller import ClusterEnvController
from anyscale.sdk.anyscale_client import (
    CreateBYODClusterEnvironment,
    CreateClusterEnvironment,
)
from frontend.cli.anyscale.utils.ray_version_utils import (
    get_correct_name_for_base_image_name,
)


def get_imported_function_name(func: Callable[..., Any]) -> str:
    return ".".join([ClusterEnvController.__module__, func.__name__])


@pytest.mark.parametrize("cluster_env_name", ["mock_cluster_env_name", None])
@pytest.mark.parametrize("build_id", ["build_id_from_cli", None])
def test_get(
    mock_auth_api_client, cluster_env_name: Optional[str], build_id: Optional[str]
) -> None:
    cluster_env_controller = ClusterEnvController()
    example_cluster_env_config = '{ "foo": "bar" }'
    cluster_env_controller.api_client.get_build_api_v2_builds_build_id_get = Mock(
        return_value=Mock(result=Mock(config_json=example_cluster_env_config))
    )
    mock_build_id_from_cluster_env = "build_id_from_cluster_env"
    possible_error_if_both_or_neither_arg_provided = (
        pytest.raises(ClickException)
        if bool(cluster_env_name) == bool(build_id)
        else nullcontext()
    )
    with patch(
        get_imported_function_name(get_build_from_cluster_env_identifier),
        new=Mock(return_value=Mock(id=mock_build_id_from_cluster_env)),
    ) as mock_get_build_from_cluster_env_identifier, possible_error_if_both_or_neither_arg_provided as e:
        cluster_env_controller.get(cluster_env_name, build_id)
    if not e:
        if cluster_env_name:
            mock_get_build_from_cluster_env_identifier.assert_called_with(
                cluster_env_name, cluster_env_controller.anyscale_api_client
            )
            mock_build_id = mock_build_id_from_cluster_env
        else:
            mock_get_build_from_cluster_env_identifier.assert_not_called()
            mock_build_id = build_id  # type: ignore
        cluster_env_controller.api_client.get_build_api_v2_builds_build_id_get.assert_called_once_with(
            mock_build_id
        )
    else:
        assert "Please only provide one of" in e.value.message
        mock_get_build_from_cluster_env_identifier.assert_not_called()
        cluster_env_controller.api_client.get_build_api_v2_builds_build_id_get.assert_not_called()


def test_list_builds(mock_auth_api_client,):
    with patch.multiple(
        "anyscale.authenticate.AuthenticationBlock",
        _validate_api_client_auth=Mock(),
        _validate_credentials_format=Mock(),
    ):
        cluster_env_controller = ClusterEnvController()
    mock_list_builds = Mock(
        return_value=[
            Mock(
                id="mock_build1",
                revision="mock_revision",
                status="mock_status",
                last_modified_at=datetime.now(),
            )
        ]
    )
    cluster_env_controller.anyscale_api_client.get_cluster_environment = Mock(
        return_value=Mock(result=Mock(name="mock_cluster_env_name"))
    )
    mock_get_cluster_env_from_name = Mock(return_value=Mock(id="mock_cluster_env_id"))

    with patch.multiple(
        "anyscale.controllers.cluster_env_controller",
        list_builds=mock_list_builds,
        get_cluster_env_from_name=mock_get_cluster_env_from_name,
    ):
        cluster_env_controller._list_builds("mock_cluster_env_name", None, max_items=20)
        mock_get_cluster_env_from_name.assert_called_with(
            "mock_cluster_env_name", cluster_env_controller.anyscale_api_client
        )
        mock_list_builds.assert_called_with(
            "mock_cluster_env_id",
            cluster_env_controller.anyscale_api_client,
            max_items=20,
        )

        cluster_env_controller._list_builds(None, "mock_cluster_env_id", max_items=20)
        mock_list_builds.assert_called_with(
            "mock_cluster_env_id",
            cluster_env_controller.anyscale_api_client,
            max_items=20,
        )
        cluster_env_controller.anyscale_api_client.get_cluster_environment.assert_called_with(
            "mock_cluster_env_id"
        )


@pytest.mark.parametrize("include_shared", [True, False])
def test_list_cluster_envs(
    mock_auth_api_client, include_shared: bool,
):
    cluster_env_controller = ClusterEnvController()
    mock_cluster_env = Mock()
    mock_cluster_env.name = "mock_cluster_env_name"
    cluster_env_controller.anyscale_api_client.search_cluster_environments = Mock(
        return_value=Mock(
            results=[mock_cluster_env], metadata=Mock(next_paging_token=None)
        )
    )
    cluster_env_controller.api_client.get_user_info_api_v2_userinfo_get = Mock(
        return_value=Mock(result=Mock(id="mock_user_id"))
    )
    mock_get_build_from_cluster_env_identifier = Mock(
        return_value=Mock(id="mock_build_id")
    )
    cluster_env_controller.anyscale_api_client.get_cluster_environment_build = Mock(
        return_value=Mock(result=Mock(last_modified_at=datetime.now()))
    )

    with patch.multiple(
        "anyscale.controllers.cluster_env_controller",
        get_build_from_cluster_env_identifier=mock_get_build_from_cluster_env_identifier,
    ):
        cluster_env_controller._list_cluster_envs(include_shared, max_items=20)
        if not include_shared:
            cluster_env_controller.anyscale_api_client.search_cluster_environments.assert_called_with(
                {"creator_id": "mock_user_id", "paging": {"count": 20}}
            )
        else:
            cluster_env_controller.anyscale_api_client.search_cluster_environments.assert_called_with(
                {"creator_id": None, "paging": {"count": 20}}
            )
        mock_get_build_from_cluster_env_identifier.assert_called_with(
            "mock_cluster_env_name", cluster_env_controller.anyscale_api_client
        )


def test_disambiguate_byod_builds(mock_auth_api_client):
    # Test that CLI can disambiguate which type of build to create based on
    # the contents of cluster-env yaml
    build_cluster_environment_mock = Mock()
    mock_anyscale_sdk = Mock(build_cluster_environment=build_cluster_environment_mock)
    cluster_env_controller = ClusterEnvController(
        anyscale_sdk=mock_anyscale_sdk, initialize_auth_api_client=False,
    )
    byod_config = {"docker_image": "a.b.c/d:efg", "ray_version": "nightly"}
    expected_byod_create_build = CreateBYODClusterEnvironment(
        name="byod", config_json=byod_config,
    )

    non_byod_config = {"env_vars": {}}
    expected_non_byod_create_build = CreateClusterEnvironment(
        name="non-byod", config_json=non_byod_config,
    )

    with tempfile.NamedTemporaryFile("wt") as f:
        yaml.dump(byod_config, f)
        f.flush()
        cluster_env_controller.build("byod", f.name)
        build_cluster_environment_mock.assert_called_with(
            expected_byod_create_build, log_output=True
        )

    with tempfile.NamedTemporaryFile("wt") as f:
        yaml.dump(non_byod_config, f)
        f.flush()
        cluster_env_controller.build("non-byod", f.name)
        build_cluster_environment_mock.assert_called_with(
            expected_non_byod_create_build, log_output=True
        )


@pytest.mark.parametrize(
    ("base_image_name", "expected_result"),
    [
        ("anyscale/ray:2.8.0-py38-cuda121", "anyscale/ray:2.8.0-py38-cuda121"),
        ("anyscale/ray:2.7.1oss-py38-cuda121", "anyscale/ray:2.7.1oss-py38-cuda121"),
        (
            "anyscale/ray:2.7.1optimized-py38-cuda121",
            "anyscale/ray:2.7.1optimized-py38-cuda121",
        ),
        ("anyscale/ray:2.7.1-py38-cuda121", "anyscale/ray:2.7.1optimized-py38-cuda121"),
        ("anyscale/ray:2.7.0-py38-cuda121", "anyscale/ray:2.7.0optimized-py38-cuda121"),
        ("anyscale/ray:2.6.0-py38", "anyscale/ray:2.6.0-py38"),
        ("anyscale/ray:2.6.0", "anyscale/ray:2.6.0"),
        ("anyscale/ray:2.1.1rc1", "anyscale/ray:2.1.1rc1"),
        ("anyscale/ray:2.0.1rc-py38-cuda121", "anyscale/ray:2.0.1rc-py38-cuda121"),
    ],
)
def test_get_correct_name_for_base_image_name(base_image_name, expected_result):
    result = get_correct_name_for_base_image_name(base_image_name)
    assert result == expected_result



import os
from typing import Optional

import click
import requests

from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client.models.create_fine_tuning_job_product_request import (
    CreateFineTuningJobProductRequest,
)
from anyscale.client.openapi_client.models.presigned_upload_file import (
    PresignedUploadFile,
)
from anyscale.client.openapi_client.models.presigned_upload_file_request import (
    PresignedUploadFileRequest,
)
from anyscale.controllers.base_controller import BaseController
from anyscale.models.fine_tune_model import FineTuneConfig
from anyscale.util import get_endpoint


class FineTuneController(BaseController):
    def __init__(
        self,
        log: Optional[LogsLogger] = None,
        initialize_auth_api_client: bool = True,
        raise_structured_exception: bool = False,
    ):
        if log is None:
            log = LogsLogger()

        super().__init__(initialize_auth_api_client, raise_structured_exception)
        self.log = log
        self.log.open_block("Output")
        self._cluster_journal_events_start_line = 0

    def _get_presigned_upload_file(
        self, file_path: str, cloud_id: str
    ) -> PresignedUploadFile:
        train_filename = os.path.basename(file_path)
        presigned_upload_file_request = PresignedUploadFileRequest(
            filename=train_filename, cloud_id=cloud_id
        )

        presigned_upload_file = self.api_client.get_file_upload_url_api_v2_files_upload_url_post(
            presigned_upload_file_request
        )

        train_file_upload_url = presigned_upload_file.presigned_url

        resolved_file_path = os.path.expanduser(file_path)
        with open(resolved_file_path, "rb") as file:
            response = requests.put(train_file_upload_url, data=file)
            if response.status_code != 200:
                raise click.ClickException(f"File upload for {file_path} failed.")

        self.log.info(f"Successfully uploaded the {file_path} to the cloud storage.")

        return presigned_upload_file

    def fine_tune(
        self, config: FineTuneConfig,
    ):
        """
        This method will upload the files to the cloud storage and then trigger the fine-tuning job.
        """
        if not os.path.isfile(config.train_file):
            self.log.error(f"Please confirm that {config.train_file} is a valid path.")
            return
        if config.valid_file and not os.path.isfile(config.valid_file):
            self.log.error(f"Please confirm that {config.valid_file} is a valid path.")
            return

        presigned_upload_train_file = self._get_presigned_upload_file(
            config.train_file, cloud_id=config.cloud_id
        )
        if config.valid_file:
            presigned_upload_valid_file = self._get_presigned_upload_file(
                config.valid_file, cloud_id=config.cloud_id
            )
            validation_file = presigned_upload_valid_file.file_id
        else:
            validation_file = None

        create_fine_tuning_job_request = CreateFineTuningJobProductRequest(
            training_file=presigned_upload_train_file.file_id,
            validation_file=validation_file,
            model=config.base_model,
            suffix=config.suffix,
            cloud_id=config.cloud_id,
            version=config.version,
        )
        ft_job = self.api_client.create_fine_tuning_job_api_v2_fine_tuning_jobs_create_post(
            create_fine_tuning_job_request
        ).result
        self.log.info(
            f'View the job in the UI at {get_endpoint(f"/fine-tuning/{ft_job.id}")}'
        )



import os
from typing import Optional
from unittest.mock import ANY, Mock, mock_open as mock_open_factory, patch

import click
import pytest

from anyscale.connect_utils.project import ProjectBlock
from anyscale.sdk.anyscale_client import ComputeTemplateQuery


@pytest.fixture()
def mock_project_block():
    with patch.multiple(
        "anyscale.connect_utils.project.ProjectBlock", __init__=Mock(return_value=None)
    ):
        project_block = ProjectBlock()
        project_block.anyscale_api_client = Mock()
        project_block.api_client = Mock()
        project_block.log = Mock()
        project_block.block_label = ""
        return project_block


@pytest.mark.parametrize("parent_cloud_id", ["mock_parent_cloud_id", None])
@pytest.mark.parametrize("cloud_isolation_ff_on", [True, False])
def test_get_default_project(
    mock_project_block, parent_cloud_id: Optional[str], cloud_isolation_ff_on: bool
):
    mock_project = Mock(id="project_id")
    mock_project.name = "project_name"
    mock_project_block.anyscale_api_client.get_default_project = Mock(
        return_value=Mock(result=mock_project)
    )
    mock_project_block.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
        return_value=Mock(result=Mock(is_on=cloud_isolation_ff_on))
    )

    assert mock_project_block._get_default_project(parent_cloud_id) == (
        mock_project.id,
        mock_project.name,
    )
    mock_project_block.anyscale_api_client.get_default_project.assert_called_once_with(
        parent_cloud_id=parent_cloud_id
    )


@pytest.mark.parametrize("cloud_isolation_ff_on", [True, False])
@pytest.mark.parametrize("cloud_name", [None, "mock_cloud_name"])
@pytest.mark.parametrize("cluster_compute_name", [None, "mock_cluster_compute_name"])
def test_get_parent_cloud_id(
    mock_project_block,
    cloud_isolation_ff_on: bool,
    cloud_name: str,
    cluster_compute_name: str,
):
    mock_get_cloud_id_and_name = Mock(return_value=("mock_parent_cloud_id", ANY))
    mock_project_block.api_client.search_compute_templates_api_v2_compute_templates_search_post = Mock(
        return_value=Mock(results=[Mock(config=Mock(cloud_id="mock_parent_cloud_id"))])
    )
    mock_get_organization_default_cloud = Mock(return_value=None)
    mock_get_last_used_cloud = Mock(return_value="mock_parent_cloud_name")

    with patch.multiple(
        "anyscale.connect_utils.project",
        get_cloud_id_and_name=mock_get_cloud_id_and_name,
        get_organization_default_cloud=mock_get_organization_default_cloud,
        get_last_used_cloud=mock_get_last_used_cloud,
    ):
        assert (
            mock_project_block._get_parent_cloud_id(
                cloud_name, cluster_compute_name, None
            )
            == "mock_parent_cloud_id"
        )
    if cloud_name:
        mock_get_cloud_id_and_name.assert_called_once_with(
            api_client=mock_project_block.api_client,
            cloud_id=None,
            cloud_name=cloud_name,
        )
    elif cluster_compute_name:
        mock_project_block.api_client.search_compute_templates_api_v2_compute_templates_search_post.assert_called_once_with(
            ComputeTemplateQuery(
                orgwide=True,
                name={"equals": cluster_compute_name},
                include_anonymous=True,
            )
        )
    else:
        mock_get_organization_default_cloud.assert_called_once_with(
            mock_project_block.api_client
        )
        mock_get_last_used_cloud.assert_called_once_with(
            None, mock_project_block.anyscale_api_client
        )
        mock_get_cloud_id_and_name.assert_called_once_with(
            mock_project_block.api_client,
            cloud_id=None,
            cloud_name="mock_parent_cloud_name",
        )


def test_ensure_project_setup_at_dir_config_exists(mock_project_block):
    # Case where .anyscale.yaml exists and is correctly set up
    with patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.project", get_project_id=Mock(return_value="project_id")
    ), patch("builtins.open", mock_open_factory()):
        mock_project = Mock(id="project_id")
        mock_project.name = "project_name"
        mock_project_block.anyscale_api_client.get_project = Mock(
            return_value=Mock(result=mock_project)
        )
        assert mock_project_block._ensure_project_setup_at_dir(
            "mock_project_dir", None
        ) == ("project_id", "project_name")

    # Case where .anyscale.yaml exists but has invalid project id
    with patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.project", get_project_id=Mock(return_value="project_id")
    ), patch("builtins.open", mock_open_factory()):
        mock_project = Mock(id="project_id")
        mock_project.name = "project_name"
        mock_project_block.anyscale_api_client.get_project = Mock(
            side_effect=click.ClickException("")
        )
        with pytest.raises(click.ClickException):
            assert mock_project_block._ensure_project_setup_at_dir(
                "mock_project_dir", None
            )
        mock_project_block.anyscale_api_client.get_project.assert_called_once_with(
            "project_id"
        )

    # Case where .anyscale.yaml exists but doesn't have project_id in the file
    with patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.project", get_project_id=Mock(return_value=None),
    ), patch("builtins.open", mock_open_factory()), pytest.raises(click.ClickException):
        assert mock_project_block._ensure_project_setup_at_dir("mock_project_dir", None)


@pytest.mark.parametrize("project_id", [None, "project_id"])
@pytest.mark.parametrize("parent_cloud_id", [None, "mock_parent_cloud_id"])
@pytest.mark.parametrize("cloud_isolation_ff_on", [True, False])
def test_ensure_project_setup_at_dir_no_config(
    mock_project_block,
    project_id: Optional[str],
    parent_cloud_id: Optional[str],
    cloud_isolation_ff_on: bool,
):
    # Case where .anyscale.yaml doesn't exist

    mock_project_block.anyscale_api_client.search_projects = Mock(
        return_value=Mock(results=[Mock(id=project_id)])
    )
    mock_project_block.anyscale_api_client.create_project = Mock(
        return_value=Mock(result=Mock(id="project_id"))
    )
    mock_project_block.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
        return_value=Mock(result=Mock(is_on=cloud_isolation_ff_on))
    )
    mock_yaml_dump = Mock()
    with patch.multiple("os.path", exists=Mock(return_value=False)), patch(
        "builtins.open", mock_open_factory()
    ) as mock_open, patch.multiple("yaml", dump=mock_yaml_dump):
        mock_project_block._ensure_project_setup_at_dir(
            "mock_project_dir", "project_name", parent_cloud_id
        )
    if not project_id:
        mock_project_block.anyscale_api_client.create_project.assert_called_once_with(
            {
                "name": "project_name",
                "description": "Automatically created by Anyscale Connect",
                "parent_cloud_id": parent_cloud_id,
            }
        )
    mock_open.assert_called_once_with("mock_project_dir/.anyscale.yaml", "w+")
    mock_yaml_dump.assert_called_once_with({"project_id": "project_id"})


@pytest.mark.parametrize("project_exists", [True, False])
@pytest.mark.parametrize("parent_cloud_id", [None, "mock_parent_cloud_id"])
@pytest.mark.parametrize("cloud_isolation_ff_on", [True, False])
def test_create_or_get_project_from_name(
    mock_project_block,
    project_exists: bool,
    parent_cloud_id: Optional[str],
    cloud_isolation_ff_on: bool,
):
    mock_find_project_id = Mock(
        return_value="test_project_id" if project_exists else None
    )
    mock_project_block.anyscale_api_client.create_project = Mock(
        return_value=Mock(result=Mock(id="test_project_id"))
    )
    mock_project_block.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
        return_value=Mock(result=Mock(is_on=cloud_isolation_ff_on))
    )

    with patch.multiple(
        "anyscale.connect_utils.project", find_project_id=mock_find_project_id
    ):
        assert mock_project_block._create_or_get_project_from_name(
            "test_project_name", parent_cloud_id
        ) == ("test_project_id", "test_project_name")

    if not project_exists:
        mock_project_block.anyscale_api_client.create_project.assert_called_once_with(
            {
                "name": "test_project_name",
                "description": "Automatically created by Anyscale Connect",
                "parent_cloud_id": parent_cloud_id,
            }
        )
    else:
        mock_project_block.anyscale_api_client.create_project.assert_not_called()


@pytest.mark.parametrize("project_dir", [None, "test_project_dir"])
@pytest.mark.parametrize("project_name", [None, "test_project_name"])
@pytest.mark.parametrize("cloud_name", [None, "test_cloud_name"])
@pytest.mark.parametrize("compute_template_name", [None, "test_compute_template_name"])
@pytest.mark.parametrize("project_exists_in_dir_tree", [True, False])
def test_init(
    project_dir: Optional[str],
    project_exists_in_dir_tree: bool,
    project_name: Optional[str],
    cloud_name: Optional[str],
    compute_template_name: Optional[str],
):
    mock_get_default_project = Mock(
        return_value=("test_default_project_id", "test_default_project_name")
    )
    mock_ensure_project_setup_at_dir = Mock(
        return_value=("test_project_id", "test_project_name")
    )
    mock_find_project_root = Mock(
        return_value="test_project_dir" if project_exists_in_dir_tree else None
    )
    mock_create_or_get_project_from_name = Mock(
        return_value=("test_project_id", "test_project_name")
    )
    mock_get_parent_cloud_id = Mock(return_value="mock_parent_cloud_id")
    with patch.multiple(
        "anyscale.connect_utils.project.ProjectBlock",
        _get_default_project=mock_get_default_project,
        _ensure_project_setup_at_dir=mock_ensure_project_setup_at_dir,
        _create_or_get_project_from_name=mock_create_or_get_project_from_name,
        _get_parent_cloud_id=mock_get_parent_cloud_id,
    ), patch.multiple(
        "anyscale.connect_utils.project", get_auth_api_client=Mock(return_value=Mock())
    ), patch.multiple(
        "anyscale.project", find_project_root=mock_find_project_root
    ):
        project_block = ProjectBlock(
            project_dir=project_dir,
            project_name=project_name,
            cloud_name=cloud_name,
            cluster_compute_name=compute_template_name,
        )
        if project_dir:
            # Project dir passed in
            assert project_block.project_dir == os.path.abspath(
                os.path.expanduser(project_dir)
            )
            assert project_block.project_name == "test_project_name"
            assert project_block.project_id == "test_project_id"
            mock_ensure_project_setup_at_dir.assert_called_once_with(
                os.path.abspath(os.path.expanduser(project_dir)),
                project_name,
                "mock_parent_cloud_id",
            )
        elif project_exists_in_dir_tree and not project_name:
            # Find project dir in local directory structure
            assert project_block.project_dir == "test_project_dir"
            assert project_block.project_name == "test_project_name"
            assert project_block.project_id == "test_project_id"
            mock_ensure_project_setup_at_dir.assert_called_once_with(
                "test_project_dir", project_name, "mock_parent_cloud_id"
            )
        elif project_name:
            assert project_block.project_dir is None
            assert project_block.project_name == "test_project_name"
            assert project_block.project_id == "test_project_id"
            mock_create_or_get_project_from_name.assert_called_once_with(
                "test_project_name", "mock_parent_cloud_id"
            )
        else:
            # Use default project
            assert project_block.project_dir is None
            assert project_block.project_name == "test_default_project_name"
            assert project_block.project_id == "test_default_project_id"
            mock_get_default_project.assert_called_once_with("mock_parent_cloud_id")



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class BillingInformation(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'edit_payment_profiles': 'str',
        'aioa_billing_information': 'ServiceUsage',
        'aica_billing_information': 'ServiceUsage'
    }

    attribute_map = {
        'edit_payment_profiles': 'edit_payment_profiles',
        'aioa_billing_information': 'aioa_billing_information',
        'aica_billing_information': 'aica_billing_information'
    }

    def __init__(self, edit_payment_profiles=None, aioa_billing_information=None, aica_billing_information=None, local_vars_configuration=None):  # noqa: E501
        """BillingInformation - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._edit_payment_profiles = None
        self._aioa_billing_information = None
        self._aica_billing_information = None
        self.discriminator = None

        self.edit_payment_profiles = edit_payment_profiles
        self.aioa_billing_information = aioa_billing_information
        self.aica_billing_information = aica_billing_information

    @property
    def edit_payment_profiles(self):
        """Gets the edit_payment_profiles of this BillingInformation.  # noqa: E501

        Link to edit payment information for organization.  # noqa: E501

        :return: The edit_payment_profiles of this BillingInformation.  # noqa: E501
        :rtype: str
        """
        return self._edit_payment_profiles

    @edit_payment_profiles.setter
    def edit_payment_profiles(self, edit_payment_profiles):
        """Sets the edit_payment_profiles of this BillingInformation.

        Link to edit payment information for organization.  # noqa: E501

        :param edit_payment_profiles: The edit_payment_profiles of this BillingInformation.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and edit_payment_profiles is None:  # noqa: E501
            raise ValueError("Invalid value for `edit_payment_profiles`, must not be `None`")  # noqa: E501

        self._edit_payment_profiles = edit_payment_profiles

    @property
    def aioa_billing_information(self):
        """Gets the aioa_billing_information of this BillingInformation.  # noqa: E501

        Billing information for AIOA subscription.  # noqa: E501

        :return: The aioa_billing_information of this BillingInformation.  # noqa: E501
        :rtype: ServiceUsage
        """
        return self._aioa_billing_information

    @aioa_billing_information.setter
    def aioa_billing_information(self, aioa_billing_information):
        """Sets the aioa_billing_information of this BillingInformation.

        Billing information for AIOA subscription.  # noqa: E501

        :param aioa_billing_information: The aioa_billing_information of this BillingInformation.  # noqa: E501
        :type: ServiceUsage
        """
        if self.local_vars_configuration.client_side_validation and aioa_billing_information is None:  # noqa: E501
            raise ValueError("Invalid value for `aioa_billing_information`, must not be `None`")  # noqa: E501

        self._aioa_billing_information = aioa_billing_information

    @property
    def aica_billing_information(self):
        """Gets the aica_billing_information of this BillingInformation.  # noqa: E501

        Billing information for AICA subscription.  # noqa: E501

        :return: The aica_billing_information of this BillingInformation.  # noqa: E501
        :rtype: ServiceUsage
        """
        return self._aica_billing_information

    @aica_billing_information.setter
    def aica_billing_information(self, aica_billing_information):
        """Sets the aica_billing_information of this BillingInformation.

        Billing information for AICA subscription.  # noqa: E501

        :param aica_billing_information: The aica_billing_information of this BillingInformation.  # noqa: E501
        :type: ServiceUsage
        """
        if self.local_vars_configuration.client_side_validation and aica_billing_information is None:  # noqa: E501
            raise ValueError("Invalid value for `aica_billing_information`, must not be `None`")  # noqa: E501

        self._aica_billing_information = aica_billing_information

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, BillingInformation):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, BillingInformation):
            return True

        return self.to_dict() != other.to_dict()



import json
import re
import secrets
import time
from typing import Dict, List, Optional, Set, Tuple

from botocore.exceptions import ClientError, NoCredentialsError
from click import ClickException

from anyscale import __version__ as anyscale_version
from anyscale.anyscale_pydantic import BaseModel
from anyscale.aws_iam_policies import (
    ANYSCALE_IAM_POLICIES,
    AnyscaleIAMPolicy,
)
from anyscale.cli_logger import CloudSetupLogger, pad_string
from anyscale.util import (
    _client,
    confirm,
    generate_inline_policy_parameter,
    generate_inline_policy_resource,
    get_anyscale_cross_account_iam_policies,
    MEMORY_DB_OUTPUT,
    MEMORY_DB_RESOURCE,
)
from anyscale.utils.cloud_utils import modify_memorydb_parameter_group


DETECT_DRIFT_TIMEOUT_SECONDS = 60 * 5  # 5 minutes
CREATE_CHANGE_SET_TIMEOUT_SECONDS = 60 * 5  # 5 minutes
UPDATE_CLOUDFORMATION_STACK_TIMEOUT_SECONDS = 60 * 5  # 5 minutes
# Some resources (e.g. memoryDB) take a long time to create, so we need to increase the timeout.
CLOUDFORMATION_TIMEOUT_SECONDS_LONG = 1800  # 30 minutes


CUSTOMER_DRIFTS_POLICY_NAME = "Customer_Drifts_Policy"
MEMORYDB_REDIS_PORT = "6379"


class PropertyDifference(BaseModel):
    """
    PropertyDifference is a field in the drift response.
    It describes the difference between the expected and actual value of a property.

    Example:
    {
        "DifferenceType": "NOT_EQUAL",
        "PropertyPath": "/Policies/0/PolicyDocument/Statement/0/Sid",
        "ExpectedValue": "AllowAnyScaleCLI",
        "ActualValue": "AllowAnyScaleCLI2"
    }

    We use the PropertyPath to recognize the drift type and identify the policy and statement.
    """

    DifferenceType: str
    PropertyPath: str
    ExpectedValue: Optional[str]
    ActualValue: Optional[str]

    def get_policy_number(self) -> Optional[int]:
        """
        Get the policy number from the PropertyPath.
        If the PropertyPath is not a policy path, return None.
        """
        match = re.search(r"\/Policies\/(\d+)", self.PropertyPath)
        if match is None:
            return None
        return int(match.group(1))

    def get_statement_number(self) -> Optional[int]:
        """
        Get the statement number from the PropertyPath.
        If the PropertyPath is not a statement path, return None.
        """
        match = re.search(r"\/Statement\/(\d+)", self.PropertyPath)
        if match is None:
            return None
        return int(match.group(1))

    def is_add_or_remove_statement(self) -> bool:
        """
        Check if the PropertyPath is a statement path
        and the difference type is ADD or REMOVE.
        """
        if (
            re.match(
                r"\/Policies\/(\d+)\/PolicyDocument\/Statement\/(\d+)$",
                self.PropertyPath,
            )
            is not None
        ):
            if self.DifferenceType == "ADD" or self.DifferenceType == "REMOVE":
                return True
            else:
                raise ClickException(f"Drift {self.json()} cannot be resolved.")
        return False


class AWSCloudformationHandler:
    def __init__(
        self, aws_cloudformation_stack_id: str, region: str, logger: CloudSetupLogger
    ):
        self.aws_cloudformation_stack_id = aws_cloudformation_stack_id
        self.region = region
        self.logger = logger
        self.cfn_client = _client("cloudformation", self.region)
        try:
            cfn_stacks = self.cfn_client.describe_stacks(
                StackName=self.aws_cloudformation_stack_id
            )["Stacks"]
        except (ClientError, NoCredentialsError) as e:
            if isinstance(e, NoCredentialsError):
                raise ClickException(
                    "Unable to locate AWS credentials. Please make sure you have AWS credentials configured.",
                )
            raise ClickException(
                f"Failed to describe cloudformation stack {self.aws_cloudformation_stack_id}. Error: {e}"
            )

        if len(cfn_stacks) != 1:
            raise ClickException(
                f"Got unexpected number of cloudformation stacks returned with stack ID: {self.aws_cloudformation_stack_id}.\n "
                f"Results returned: {cfn_stacks}\n"
                "Please check your cloudformation stack page to make sure the stack still exists."
            )
        cfn_stack = cfn_stacks[0]

        # Validate the cloudformation stack status
        if not cfn_stack["StackStatus"].endswith("_COMPLETE"):
            raise ClickException(
                f"Cloudformation stack {cfn_stack['StackName']} is in {cfn_stack['StackStatus']} status. Please check your cloudformation stack page to make sure the stack is in a stable state."
            )

        self.stack = cfn_stack

    def get_cloudformation_stack(self):
        return self.stack

    def get_resource(self, logical_resource_id: str):
        try:
            resource = self.cfn_client.describe_stack_resource(
                StackName=self.aws_cloudformation_stack_id,
                LogicalResourceId=logical_resource_id,
            )
            return resource["StackResourceDetail"]
        except ClientError:
            raise ClickException(
                f"Failed to get resource {logical_resource_id} from stack {self.aws_cloudformation_stack_id}."
            )

    def detect_drift(self) -> List[Dict]:
        """
        Detect drifts on cloudformation stack.
        More about drifts on cfn stack: https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/using-cfn-stack-drift.html
        """
        stack_name = self.aws_cloudformation_stack_id
        with self.logger.spinner("Detecting drift on cloudformation stack..."):
            # Init drift detection
            try:
                drift_detection_id = self.cfn_client.detect_stack_drift(
                    StackName=stack_name
                )["StackDriftDetectionId"]
            except ClientError as e:
                raise ClickException(
                    f"Failed to detect drift on stack {stack_name}: {e}"
                )

            # Polling drift detection status
            end_time = time.time() + DETECT_DRIFT_TIMEOUT_SECONDS
            while time.time() < end_time:
                time.sleep(1)
                try:
                    response = self.cfn_client.describe_stack_drift_detection_status(
                        StackDriftDetectionId=drift_detection_id
                    )
                except ClientError as e:
                    raise ClickException(
                        f"Failed to get drift detection status for stack {stack_name}: {e}"
                    )
                if response["DetectionStatus"] == "DETECTION_COMPLETE":
                    drift_response = self.cfn_client.describe_stack_resource_drifts(
                        StackName=stack_name,
                        StackResourceDriftStatusFilters=[
                            "MODIFIED",
                            "DELETED",
                            "NOT_CHECKED",
                        ],
                    )
                    drifts = drift_response["StackResourceDrifts"]
                    self.logger.info("Drift detection completed.")
                    return drifts
                elif response["DetectionStatus"] == "DETECTION_FAILED":
                    raise ClickException(
                        f'Drift detection failed. Error: {response["DetectionStatusReason"]}'
                    )
        raise ClickException("Drift detection timeout. Please try again later.")

    def resolve_drift(self, drift: Dict):
        """
        Resolve the drift on the cross account IAM role.

        1) Drifts we care about:
        We only resolve the drifts on the Anyscale-maintained inline policies (defined in ANYSCALE_IAM_POLICIES).
        For other drifts, we'll skip them and keep the drifts as is.

        2) How to resolve the drift:
        We first identify the drifted statements in the Anyscale-maintained inline policies from the drift detection response.
        Then we append the drifted statements to the Customer_Drifts_Policy of the cross account IAM role.
        The reason is that we'll overwrite the Anyscale-maintained inline policies and we don't want to lose the drifted statements.
        If the customer removes some of the statements, they'll need to remove it again manually.
        """
        try:
            drifted_statements = extract_drifted_statements(drift)
            if len(drifted_statements) == 0:
                self.logger.info("No drifted statements found.")
                return True
            role_name = drift["PhysicalResourceId"]
            append_statements_to_customer_drifts_policy(
                self.region, role_name, drifted_statements
            )
            self.logger.info(
                f"Drifted statements have been appended to the policy {CUSTOMER_DRIFTS_POLICY_NAME} of the role {role_name}."
            )
            return True
        except Exception as e:  # noqa: BLE001
            self.logger.error(f"Failed to resolve drift. Error: {e}")
            return False

    def get_template_body(self):
        try:
            template_body = self.cfn_client.get_template(
                StackName=self.aws_cloudformation_stack_id, TemplateStage="Original"
            )["TemplateBody"]
            return template_body
        except ClientError as e:
            raise ClickException(
                f"Failed to get template body for stack {self.aws_cloudformation_stack_id}: {e}"
            )

    def update_cloudformation_stack(
        self,
        updated_template_body: str,
        updated_parameters: List[Dict],
        yes: bool = False,
        timeout_seconds: int = UPDATE_CLOUDFORMATION_STACK_TIMEOUT_SECONDS,
    ):
        """
        Update the cloudformation stack with the updated parameter list and template body.
        Update the stack by creating a change set and executing it.
        """
        stack_name = self.aws_cloudformation_stack_id

        # Create change set
        with self.logger.spinner("Creating change set for cloud update..."):
            response = self.cfn_client.create_change_set(
                StackName=stack_name,
                ChangeSetName=f"AnyscaleCloudUpdate{str(secrets.token_hex(4))}",
                TemplateBody=updated_template_body,
                Parameters=updated_parameters,
                Capabilities=["CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND"],
                ChangeSetType="UPDATE",
            )

            change_set_id = response["Id"]

            # Polling change set status
            end_time = time.time() + CREATE_CHANGE_SET_TIMEOUT_SECONDS
            while time.time() < end_time:
                time.sleep(1)
                response = self.cfn_client.describe_change_set(
                    ChangeSetName=change_set_id, StackName=stack_name
                )
                if response["Status"] == "CREATE_COMPLETE":
                    break
                elif response["Status"] == "FAILED":
                    self.cfn_client.delete_change_set(ChangeSetName=change_set_id)
                    raise ClickException(
                        f"Failed to create change set for cloud update. {response['StatusReason']}"
                    )
            else:
                raise ClickException(
                    "Timeout when creating change set for cloud update. Please try again later."
                )

        # Preview change set
        stack_id = response["StackId"]
        stack_url = f"https://{self.region}.console.aws.amazon.com/cloudformation/home?region={self.region}#/stacks/stackinfo?stackId={stack_id}"
        change_set_url = f"https://{self.region}.console.aws.amazon.com/cloudformation/home?region={self.region}#/stacks/changesets/changes?stackId={stack_id}&changeSetId={change_set_id}"
        self.logger.info(f"Change set created at {change_set_url}")
        confirm(
            "Please review the change set before updating the stack. Do you want to proceed with the update?",
            yes,
        )

        # Execute change set
        with self.logger.spinner("Executing change set for cloud update..."):
            response = self.cfn_client.execute_change_set(ChangeSetName=change_set_id)

            # Polling cfn stack status
            end_time = time.time() + timeout_seconds
            while time.time() < end_time:
                time.sleep(1)
                response = self.cfn_client.describe_stacks(StackName=stack_name)
                stack = response["Stacks"][0]
                if stack["StackStatus"] == "UPDATE_COMPLETE":
                    break
                elif stack["StackStatus"] == "UPDATE_ROLLBACK_COMPLETE":
                    raise ClickException(
                        f"Failed to execute change set. Please check the cloudformation stack events for more details ({stack_url})"
                    )
            else:
                raise ClickException(
                    f"Timeout when executing change set. Please check the cloudformation stack events for more details ({stack_url})"
                )


def validate_stack_version(stack_parameters: List[Dict]) -> bool:
    cfn_stack_version = None
    for parameter in stack_parameters:
        if parameter["ParameterKey"] == "AnyscaleCLIVersion":
            cfn_stack_version = parameter["ParameterValue"]
            break

    # return False if:
    # 1) cfn_stack_version is None
    # 2) cfn_stack_version is the different with the CLI version
    # 3) cfn_stack_version is the same as the CLI version but the CLI version is dev version
    if cfn_stack_version == anyscale_version and anyscale_version != "0.0.0-dev":
        return False
    return True


def is_template_policy_documents_up_to_date(stack_parameters: List[Dict]) -> bool:
    """
    Check if the policy documents in the cfn template are up to date.
    """
    parameter_dict = {p["ParameterKey"]: p["ParameterValue"] for p in stack_parameters}
    for policy in ANYSCALE_IAM_POLICIES:
        if policy.parameter_key not in parameter_dict:
            return False
        if parameter_dict[policy.parameter_key] != policy.policy_document:
            return False
    return True


def format_drifts(drifts: List[Dict]) -> str:
    padding_size = 40
    outputs: List[str] = []
    outputs.append(
        f'{pad_string("Resource Type", padding_size)}'
        f'{pad_string("Resource Id", padding_size)}'
        f'{pad_string("Drift status", padding_size)}'
    )
    outputs.append(
        f'{pad_string("-------------", padding_size)}'
        f'{pad_string("-----------", padding_size)}'
        f'{pad_string("------------", padding_size)}'
    )
    for drift in drifts:
        outputs.append(
            f'{pad_string(drift["ResourceType"], padding_size)}'
            f'{pad_string(drift["PhysicalResourceId"], padding_size)}'
            f'{pad_string(drift["StackResourceDriftStatus"], padding_size)}'
        )
    return "\n".join(outputs)


def extract_cross_account_iam_role_drift(drifts: List[Dict]) -> Optional[Dict]:
    """
    Check if the cross account IAM role is drifted.
    If so, return the drift information.
    """
    for drift in drifts:
        # TODO (congding): don't hardcode "customerRole"
        if (
            drift["ResourceType"] == "AWS::IAM::Role"
            and drift["LogicalResourceId"] == "customerRole"
            and drift["StackResourceDriftStatus"] != "IN_SYNC"
        ):
            return drift
    return None


def get_all_sids_from_policy(policy: Dict) -> List[str]:
    """
    Get all the SIDs from the policy.
    SIDs (Statement ID) are used to identify the statements.
    """
    sids: List[str] = []
    for statement in policy["PolicyDocument"]["Statement"]:
        if "Sid" in statement:
            sids.append(statement["Sid"])
        else:
            raise ClickException(
                f"Statement {statement} in policy {policy['PolicyName']} doesn't have a Sid."
            )
    return sids


def get_sids_to_remove(
    diffs: List[Dict], expected_policies: List[Dict],
) -> Dict[str, Set[str]]:
    """
    Get the SIDs from the drifted statements to remove from the expected policies.
    """
    policy_names_to_overwrite = [policy.policy_name for policy in ANYSCALE_IAM_POLICIES]

    sids_to_remove: Dict[str, Set[str]] = {}

    for raw_diff in diffs:
        diff = PropertyDifference(**raw_diff)
        policy_number = diff.get_policy_number()
        if policy_number is None:
            # We'll skip the drift if it's not a policy drift
            # We only update the maintained inline policies
            # If there're other drifts (e.g., AssumeRolePolicyDocument drift)
            # we'll skip them and keep the drifts as is.
            continue
        try:
            policy_name = expected_policies[policy_number]["PolicyName"]
        except IndexError:
            continue
        if policy_name not in policy_names_to_overwrite:
            # Not a policy we care about
            # We only update the maintained inline policies
            # If the drifts are on other policies (e.g., newly created inline policies),
            # we'll skip them and keep the drifts as is.
            continue

        # Drift detected on a maintained policy
        statement_number = diff.get_statement_number()
        if statement_number is None:
            raise ClickException(
                f"Drift {diff} in policy {policy_name} cannot be resolved."
            )
        if diff.is_add_or_remove_statement():
            # Policy statement added or removed
            # No need to append to the drifted policy
            continue
        expected_statements = expected_policies[policy_number]["PolicyDocument"][
            "Statement"
        ]
        try:
            sid = expected_statements[statement_number]["Sid"]
            if policy_name not in sids_to_remove:
                sids_to_remove[policy_name] = set()
            sids_to_remove[policy_name].add(sid)
        except IndexError:
            continue

    return sids_to_remove


def generate_drifted_statements_to_append(
    actual_policies: List[Dict], undrifted_sid: Dict[str, Set[str]]
) -> List[Dict]:
    """
    Generate the drifted statements to append to the drifted policy.

    Iterate through the actual policies and skip the undrifted statements.
    For the drifted statements, we'll append them to the drifted statements list.

    """
    no_sid_cnt = 0
    sid_suffix = str(secrets.token_hex(4))
    drifted_statements: List[Dict] = []

    for actual_policy in actual_policies:
        policy_name = actual_policy["PolicyName"]
        if policy_name not in undrifted_sid:
            # Not a policy we care about
            continue
        statements = actual_policy["PolicyDocument"]["Statement"]
        for statement in statements:
            sid = statement.get("Sid", None)
            if sid is None:
                # No sid in the statement
                # Generate a new sid
                statement["Sid"] = f"Drifted{no_sid_cnt}{sid_suffix}"
                no_sid_cnt += 1
                drifted_statements.append(statement)
            elif sid not in undrifted_sid[policy_name]:
                # drifted statement
                statement["Sid"] = f"{statement['Sid']}Drifted{sid_suffix}"
                drifted_statements.append(statement)

    return drifted_statements


def get_all_sids(expected_policies) -> Dict[str, Set[str]]:
    all_sids: Dict[str, Set[str]] = {}
    policies_to_overwrite = [policy.policy_name for policy in ANYSCALE_IAM_POLICIES]
    for policy in expected_policies:
        if policy["PolicyName"] not in policies_to_overwrite:
            continue
        if policy["PolicyDocument"]["Version"] != "2012-10-17":
            raise ClickException(
                f"Unexpected policy version {policy['PolicyDocument']['Version']} for policy {policy['PolicyName']}"
            )
        all_sids[policy["PolicyName"]] = set(get_all_sids_from_policy(policy))
    return all_sids


def extract_drifted_statements(drift: Dict) -> List[Dict]:  # noqa: PLR0912
    """
    Extract the drifted statements from the role drift information.

    1) About SID (Statement ID):
    Each inline policy contains a list of statements, and each statement has a SID (Statement ID).
    We use the SID to identify the statements.
    Anyscale-maintained policies is guaranteed to have a SID for each statement.

    2) Get drifted statements:
    Step 1: Get all the SIDs from the expected policies.
    Step 2: Leverage the drift detection result to identify the drifted statements.
    If there's a drift on a statement, we will add it to a set of SIDs to remove.
    Step 3: Remove the drifted SIDs and get a set of undrifted SIDs.
    Step 4: Get all the statements from the actual policy and filter out the undrfited SIDs.
    """
    # Get all sids from expected policies
    expected_policies = json.loads(drift["ExpectedProperties"])["Policies"]
    all_sids = get_all_sids(expected_policies)

    # Remove drifted statements
    diffs = drift["PropertyDifferences"]
    sids_to_remove = get_sids_to_remove(diffs, expected_policies)

    undrifted_sid: Dict[str, Set[str]] = {}
    for policy_name, sids in all_sids.items():
        if policy_name not in sids_to_remove:
            undrifted_sid[policy_name] = sids
            continue
        undrifted_sid[policy_name] = sids.difference(sids_to_remove[policy_name])

    # Get drifted statements
    actual_policies = json.loads(drift["ActualProperties"])["Policies"]
    drifted_statements = generate_drifted_statements_to_append(
        actual_policies, undrifted_sid
    )

    return drifted_statements


def append_statements_to_customer_drifts_policy(
    region: str, role_name: str, statements: List[Dict]
):
    """
    Append statements to the customer drfits policy.
    If the inline policy doesn't exist, create it.
    """
    iam = _client("iam", region)
    policy_document = None
    try:
        policy = iam.get_role_policy(
            RoleName=role_name, PolicyName=CUSTOMER_DRIFTS_POLICY_NAME,
        )
        policy_document = policy["PolicyDocument"]
    except ClientError as e:
        if e.response["Error"]["Code"] != "NoSuchEntity":
            raise e
    # Create the policy
    if policy_document is None:
        policy_document = {
            "Version": "2012-10-17",
            "Statement": [],
        }

    # Append new statements
    for statement in statements:
        policy_document["Statement"].append(statement)  # type: ignore
    try:
        iam.put_role_policy(
            RoleName=role_name,
            PolicyName=CUSTOMER_DRIFTS_POLICY_NAME,
            PolicyDocument=json.dumps(policy_document),
        )
    except ClientError as e:
        raise ClickException(
            f"Failed to append statements to the drifted policy. Error: {e}"
        )


def merge_parameters(
    existing_parameters: List[Dict], parameters_to_update: List[Dict]
) -> List[Dict]:
    """
    Overwrite the existing parameters with the parameters to update.
    If the parameter to update does not exist in the existing parameters, add it to the existing parameters.

    The returned parameter list should contain all the combined parameters.
    """
    returned_parameters: Dict = {
        p["ParameterKey"]: p["ParameterValue"] for p in existing_parameters
    }
    for p in parameters_to_update:
        returned_parameters[p["ParameterKey"]] = p["ParameterValue"]
    return [
        {"ParameterKey": k, "ParameterValue": v} for k, v in returned_parameters.items()
    ]


def add_missing_parameters_to_template_body(
    template_body: str, missing_parameters: Set[str]
) -> str:
    """
    Add missing parameters to template body.

    For AnyscaleCLIVersion, we only need to add the parameter part.

    For other parameters for inline IAM policies, we need to add both parameter and resource definitions.
    """
    # Get all the missing parameters' policy information
    policy_dict: Dict[str, AnyscaleIAMPolicy] = {}
    for policy in ANYSCALE_IAM_POLICIES:
        if policy.parameter_key in missing_parameters:
            policy_dict[policy.parameter_key] = policy

    parameter_substitutions = ["Parameters:"]
    resource_substitutions = ["Resources:"]

    for parameter_key in missing_parameters:
        if parameter_key == "AnyscaleCLIVersion":
            parameter_substitutions.append(
                "  AnyscaleCLIVersion:\n    Description: Anyscale CLI version\n    Type: String\n"
            )
        elif parameter_key == "MemoryDBRedisPort":
            parameter_substitutions.append(
                f'  MemoryDBRedisPort:\n    Description: Port for MemoryDB Redis\n    Type: String\n    Default: "{MEMORYDB_REDIS_PORT}"\n'
            )
        else:
            policy = policy_dict[parameter_key]
            parameter_substitutions.append(
                generate_inline_policy_parameter(policy) + "\n"
            )
            resource_substitutions.append(
                generate_inline_policy_resource(policy) + "\n"
            )

    template_body = template_body.replace(
        "Parameters:", "\n".join(parameter_substitutions),
    )

    template_body = template_body.replace(
        "Resources:", "\n".join(resource_substitutions),
    )
    return template_body


def try_delete_customer_drifts_policy(cloud):
    iam_client = _client("iam", cloud.region)
    role_name = cloud.credentials.split("/")[-1]
    try:
        iam_client.delete_role_policy(
            RoleName=role_name, PolicyName=CUSTOMER_DRIFTS_POLICY_NAME
        )
    except ClientError as e:
        if e.response["Error"]["Code"] != "NoSuchEntity":
            raise ClickException(
                f"Failed to delete inline policy {CUSTOMER_DRIFTS_POLICY_NAME} in role {role_name}: {e}"
            )


def update_template_with_memorydb(template_body: str, az_num: int) -> str:
    """
    Add MemoryDB resources and outputs to template body.
    """
    subnet_list = [f"        - !Ref Subnet{i}" for i in range(az_num)]
    memory_db_resource_str = MEMORY_DB_RESOURCE.format("\n".join(subnet_list))

    updated_template_body = template_body.replace(
        "Resources:", f"Resources:\n{memory_db_resource_str}\n"
    ).replace("Outputs:", f"Outputs:\n{MEMORY_DB_OUTPUT}\n")

    return updated_template_body


def generate_updated_parameters_and_template(
    cfn_stack_parameters: List[Dict],
    template_body: str,
    parameters_to_update: List[Dict],
) -> Tuple[List[Dict], str]:
    """
    Generate the updated parameter list and updated tempelate body.
    If parameter exists in both cfn_stack_parameters and parameters_to_update,
    we'll override the value in cfn_stack_parameters with the value in parameters_to_update.
    If parameter only exists in parameters_to_update, we'll add it to cfn_stack_parameters.
    We also update the template body with the missing parameters.
    """
    updated_parameters = merge_parameters(cfn_stack_parameters, parameters_to_update)
    missing_parameters = {p["ParameterKey"] for p in parameters_to_update}.difference(
        {p["ParameterKey"] for p in cfn_stack_parameters}
    )
    if len(missing_parameters) > 0:
        updated_template_body = add_missing_parameters_to_template_body(
            template_body, missing_parameters
        )
    else:
        updated_template_body = template_body
    return updated_parameters, updated_template_body


def update_iam_role(
    region: str,
    aws_cloudformation_stack_id: str,
    logger: CloudSetupLogger,
    yes: bool = False,
):
    cfn_handler = AWSCloudformationHandler(aws_cloudformation_stack_id, region, logger)
    cfn_stack = cfn_handler.get_cloudformation_stack()
    cfn_stack_parameters = cfn_stack["Parameters"]

    # Validate cfn stack version is not the same as the CLI version
    if not validate_stack_version(cfn_stack_parameters):
        logger.info(
            "Cloud is already up-to-date. Please make sure your Anyscale CLI is on the latest version."
        )
        return

    # Check if the cfn template are the same as the CLI definition
    if is_template_policy_documents_up_to_date(cfn_stack_parameters):
        logger.info("No inline policy changes detected.")
        return

    drifts = cfn_handler.detect_drift()
    role_drift = extract_cross_account_iam_role_drift(drifts)
    if role_drift:
        inline_policies = [policy.policy_name for policy in ANYSCALE_IAM_POLICIES]
        customer_role_drift_details_url = f"https://{region}.console.aws.amazon.com/cloudformation/home?region={region}#/stacks/drifts/info?stackId={aws_cloudformation_stack_id}&logicalResourceId={role_drift['LogicalResourceId']}"
        logger.info(
            f"Drfits detected on the cross account IAM role: {customer_role_drift_details_url}"
        )
        logger.warning(
            logger.highlight(
                f"Cloud update will create or override the inline policies: {inline_policies}. "
                f"We'll append drifted statements to inline policy {CUSTOMER_DRIFTS_POLICY_NAME} so you won't lose your change."
                "Please note that this will only include the permissions you added to the role. "
                "If you have removed permissions or restricted the resources from the role, "
                "you'll need to manually apply the change again after cloud update."
            )
        )
        confirm(
            "Proceed to resolve the drift?", yes,
        )
        resolved_result = cfn_handler.resolve_drift(role_drift)
        if not resolved_result:
            confirm(
                "[DANGER] If you proceed to update the cloud,"
                f"we'll overwrite the changes you made in inline policies {inline_policies}. "
                "We recommend creating a new inline policy to include your changes. "
                "Continue to overwrite the inline policies?",
                yes,
            )
    template_body = cfn_handler.get_template_body()
    # Get updated parameter list
    # We update the following 2 types of parameters:
    # 1. AnyscaleCLIVersion: the version of CLI
    # 2. Parameters that define the inline policy documents for the cross account IAM role
    # Other parameters should remain unchanged.
    parameters_to_update: List[Dict] = get_anyscale_cross_account_iam_policies()
    parameters_to_update.append(
        {"ParameterKey": "AnyscaleCLIVersion", "ParameterValue": anyscale_version}
    )

    (
        updated_parameters,
        updated_template_body,
    ) = generate_updated_parameters_and_template(
        cfn_stack_parameters, template_body, parameters_to_update
    )

    cfn_handler.update_cloudformation_stack(
        updated_template_body, updated_parameters, yes,
    )


def get_or_create_memorydb(
    region: str,
    aws_cloudformation_stack_id: str,
    az_num: int,
    logger: CloudSetupLogger,
    yes: bool = False,
) -> str:
    """
    Get or create the memorydb cluster.
    """
    cfn_handler = AWSCloudformationHandler(aws_cloudformation_stack_id, region, logger)

    cfn_stack = cfn_handler.get_cloudformation_stack()
    cfn_stack_parameters = cfn_stack["Parameters"]
    template_body = cfn_handler.get_template_body()

    if "AWS::MemoryDB::Cluster" in template_body:
        logger.info("MemoryDB cluster already exists in cloudformation template.")
        memorydb = cfn_handler.get_resource("MemoryDB")
        return memorydb["PhysicalResourceId"]

    # Update parameters
    parameters_to_update: List[Dict] = [
        {"ParameterKey": "MemoryDBRedisPort", "ParameterValue": MEMORYDB_REDIS_PORT}
    ]
    updated_parameters, template_body = generate_updated_parameters_and_template(
        cfn_stack_parameters, template_body, parameters_to_update
    )

    # Update template
    updated_template_body = update_template_with_memorydb(template_body, az_num)

    # Update cfn stack
    cfn_handler.update_cloudformation_stack(
        updated_template_body,
        updated_parameters,
        yes,
        timeout_seconds=CLOUDFORMATION_TIMEOUT_SECONDS_LONG,
    )
    memorydb_parameter_group = cfn_handler.get_resource("MemoryDBParameterGroup")
    modify_memorydb_parameter_group(
        memorydb_parameter_group["PhysicalResourceId"], region
    )

    memorydb = cfn_handler.get_resource("MemoryDB")
    return memorydb["PhysicalResourceId"]



from unittest.mock import Mock, patch

from aiohttp import ClientSession, WSMsgType
from asynctest import CoroutineMock
import pytest

from anyscale.api_utils import logs_util
from anyscale.api_utils.logs_util import (
    _download_log_from_s3_url,
    _remove_ansi_escape_sequences,
    _stream_log_from_ray_websocket,
)
from anyscale.shared_anyscale_utils.test_util import MockAsyncContextManagerReturnValue


# Indirectly tests gather_in_batches()
async def test_download_logs_concurrently():
    # Test all downloads succeed
    with patch.object(
        logs_util,
        _download_log_from_s3_url.__name__,
        CoroutineMock(side_effect=["logs_from_chunk_1\n", "logs_from_chunk_2"]),
    ) as mock_download_log_from_s3_url:
        result = await logs_util._download_logs_concurrently(
            ["http://presigned.url.1", "http://presigned.url.2"], 1
        )
    assert result == "logs_from_chunk_1\nlogs_from_chunk_2"
    assert mock_download_log_from_s3_url.call_count == 2

    # Test one download fail raise exception
    exception_msg = "Log chunk download failed"
    with patch.object(
        logs_util,
        _download_log_from_s3_url.__name__,
        CoroutineMock(side_effect=Exception(exception_msg)),
    ) as mock_download_log_from_s3_url, pytest.raises(Exception, match=exception_msg):
        result = await logs_util._download_logs_concurrently(
            ["http://presigned.url.1", "http://presigned.url.2"], 1
        )


async def test_stream_log_from_ray_websocket():
    mock_websocket = MockAsyncContextManagerReturnValue(
        receive=CoroutineMock(
            side_effect=[
                Mock(type=WSMsgType.TEXT, data="\n \x1b[1mABC "),
                Mock(type=WSMsgType.ERROR),
                Mock(type=WSMsgType.CLOSED),
            ]
        )
    )
    mock_url = "https://ray.websocket.url"
    mock_log_callback = Mock()
    # Test normal case, newlines and spaces are stripped,
    # and ANSI escape sequences are removed by default
    with patch.object(
        ClientSession, ClientSession.ws_connect.__name__, return_value=mock_websocket
    ) as mock_ws_connect:
        await _stream_log_from_ray_websocket(mock_url, mock_log_callback)
    # This is actually an `async def` fn, but is mocked as a normal `def` fn
    mock_ws_connect.assert_called_once_with(mock_url)
    assert mock_websocket.receive.call_count == 3
    mock_log_callback.assert_called_once_with("ABC")


def test_remove_ansi_escape_sequences():
    string_with_ansi_escape_sequences = "\x1b[1m\x1b[31mHello World\x1B[0m\x1B[0m"
    assert (
        _remove_ansi_escape_sequences(string_with_ansi_escape_sequences)
        == "Hello World"
    )



import os
import tempfile
from unittest.mock import Mock

import yaml

from anyscale.client.openapi_client.models import (
    SessionSshKey,
    SessionsshkeyResponse,
)
from anyscale.shared_anyscale_utils.project import CLUSTER_YAML_TEMPLATE
from anyscale.utils.workspace_utils import _get_ssh_key_info


def test_configure_ssh_key(base_mock_api_client: Mock) -> None:
    base_mock_api_client.get_session_ssh_key_api_v2_sessions_session_id_ssh_key_get = Mock(
        return_value=SessionsshkeyResponse(
            result=SessionSshKey(key_name="SSH_KEY", private_key="PRIVATE_KEY")
        )
    )
    input_config = yaml.safe_load(CLUSTER_YAML_TEMPLATE)
    input_config["head_node"] = {}
    input_config["worker_nodes"] = {}
    with tempfile.TemporaryDirectory() as directory:
        _get_ssh_key_info(input_config, "session_id", base_mock_api_client, directory)
        with open(os.path.join(directory, "SSH_KEY.pem")) as f:
            assert f.read() == "PRIVATE_KEY"



import os
from typing import List, Optional

import click
import tabulate
import yaml

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import ComputeTemplate, ProductionJob
from anyscale.client.openapi_client.models.decorated_list_service_api_model import (
    DecoratedListServiceAPIModel,
)
from anyscale.client.openapi_client.models.decorated_production_service_v2_api_model import (
    DecoratedProductionServiceV2APIModel,
)
from anyscale.client.openapi_client.models.ha_job_states import HaJobStates
from anyscale.controllers.base_controller import BaseController
from anyscale.formatters.service_formatter import format_service_config
from anyscale.models.service_model import ServiceConfig
from anyscale.project import infer_project_id
from anyscale.tables import ServicesTable
from anyscale.util import (
    get_endpoint,
    poll,
    populate_dict_with_workspace_config_if_exists,
)
from anyscale.utils.runtime_env import override_runtime_env_config
from anyscale.utils.workload_types import Workload


_POLL_SERVICE_VERSION_INFO_INTERVAL_IN_SECS = 3.0
_POLL_SERVICE_VERSION_INFO_TIMEOUT_IN_SECS = 60.0 * 3


class ServiceController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log

    def _get_services_by_name(
        self,
        *,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool,
        max_items: int,
    ) -> List[DecoratedListServiceAPIModel]:
        """Makes an API call to get all services matching the provided filters.

        This returns both v1 and v2 services.
        """
        creator_id = None
        if created_by_me:
            user_info_response = self.api_client.get_user_info_api_v2_userinfo_get()
            creator_id = user_info_response.result.id

        services_list = []
        resp = self.api_client.list_services_api_v2_services_v2_get(
            project_id=project_id, name=name, creator_id=creator_id, count=10,
        )
        services_list.extend(resp.results)
        paging_token = resp.metadata.next_paging_token
        while paging_token is not None and len(services_list) < max_items:
            resp = self.api_client.list_services_api_v2_services_v2_get(
                project_id=project_id,
                name=name,
                creator_id=creator_id,
                count=10,
                paging_token=paging_token,
            )
            services_list.extend(resp.results)
            paging_token = resp.metadata.next_paging_token

        return services_list[:max_items]

    def _get_service_id_from_name(
        self, service_name: str, project_id: Optional[str]
    ) -> str:
        """Get the ID for a service by name.

        If project_id is specified, filter to that project, else don't filter on project_id and
        instead error if there are multiple services with the name.

        Raises an exception if there are zero or multiple services with the given name.
        """
        results = self._get_services_by_name(
            name=service_name, project_id=project_id, created_by_me=False, max_items=10
        )

        if len(results) == 0:
            raise click.ClickException(
                f"No service with name '{service_name}' was found. "
                "Please verify that this service exists and you have access to it."
            )
        elif len(results) > 1:
            raise click.ClickException(
                f"There are multiple services with name '{service_name}'. "
                "Please filter using --project-id or specify the --service-id instead. "
                f"Services found: \n{ServicesTable(results)}"
            )

        return results[0].id

    def get_service_id(
        self,
        *,
        service_id: Optional[str] = None,
        service_name: Optional[str] = None,
        service_config_file: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> str:
        """Get the service ID given the ID, name, or config file.

        This is a utility used by multiple CLI commands to standardize mapping these options
        to a service_id.

        The precedence is: service_id > service_name > service_config_file.

        If the project_id is specified directly or via the service_config_file it will be used
        to filter the query. Else we will try to find the service across all projects and error
        if there are multiple.
        """
        if service_id is not None:
            if service_name is not None or service_config_file is not None:
                raise click.ClickException(
                    "Only one of service ID, name, or config file should be specified."
                )
        elif service_name is not None:
            if service_config_file is not None:
                raise click.ClickException(
                    "Only one of service ID, name, or config file should be specified."
                )
            service_id = self._get_service_id_from_name(service_name, project_id)
        elif service_config_file is not None:
            service_config: ServiceConfig = self._read_service_config_file(
                service_config_file
            )
            # Allow the passed project_id to override the one in the file.
            project_id = project_id or service_config.project_id
            service_id = self._get_service_id_from_name(service_config.name, project_id)
        else:
            raise click.ClickException(
                "Service ID, name, or config file must be specified."
            )

        return service_id

    def generate_config_from_file(
        self,
        service_config_file,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        canary_percent: Optional[int] = None,
        rollout_strategy: Optional[str] = None,
    ) -> ServiceConfig:
        """
        Given the input file path and user overrides, this method
        constructs the expected service configuration.

        Note that if the command is run from a workspace, some values
        may be overwritten.
        """
        service_config: ServiceConfig = self._read_service_config_file(
            service_config_file
        )
        if name:
            service_config.name = name

        if description:
            service_config.description = description

        if version:
            service_config.version = version

        if canary_percent is not None:
            service_config.canary_percent = canary_percent

        if rollout_strategy:
            service_config.rollout_strategy = rollout_strategy

        if service_config.rollout_strategy:
            service_config.rollout_strategy = service_config.rollout_strategy.upper()

        return service_config

    def _get_maximum_uptime_minutes(self, service: ProductionJob) -> Optional[int]:
        compute_config: ComputeTemplate = self.api_client.get_compute_template_api_v2_compute_templates_template_id_get(
            service.config.compute_config_id
        ).result
        return compute_config.config.maximum_uptime_minutes

    def _get_maximum_uptime_output(self, maximum_uptime_minutes: Optional[int]) -> str:
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return f"set to {maximum_uptime_minutes} minutes"
        return "disabled"

    def _get_additional_log_if_maximum_uptime_enabled(
        self, maximum_uptime_minutes: Optional[int]
    ) -> str:
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return " This may cause disruptions. To disable, update the compute config."
        return ""

    def _read_service_config_file(self, service_config_file: str) -> ServiceConfig:
        if not os.path.exists(service_config_file):
            raise click.ClickException(f"Config file {service_config_file} not found.")

        with open(service_config_file) as f:
            config_dict = yaml.safe_load(f)

        updated_config_dict = populate_dict_with_workspace_config_if_exists(
            config_dict, self.anyscale_api_client
        )
        return ServiceConfig.parse_obj(updated_config_dict)

    def rollout(
        self,
        service_config_file: str,
        name: Optional[str] = None,
        version: Optional[str] = None,
        canary_percent: Optional[int] = None,
        rollout_strategy: Optional[str] = None,
    ):
        """
        Deploys a Service 2.0.
        """
        config = self.generate_config_from_file(
            service_config_file,
            name=name,
            version=version,
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
        )

        config.project_id = infer_project_id(
            self.anyscale_api_client,
            self.api_client,
            self.log,
            project_id=config.project_id,
            cluster_compute_id=config.compute_config_id,
            cluster_compute=config.compute_config,
            cloud=config.cloud,
            cluster_compute_version=config.compute_config_version,
        )

        service_config = config.config
        if service_config:
            access_config = service_config.get("access", {})

            if (
                isinstance(access_config, dict)
                and access_config.get("use_bearer_token", True) is False
            ):
                self.log.warning(
                    "Using bearer token for authorization is disabled: no authorization guarding will be used for this service."
                )

        if not config.ray_serve_config:
            config.ray_serve_config = {}

        self._overwrite_runtime_env_in_v2_ray_serve_config(config)

        service_v2_config = format_service_config(config)

        service = self.api_client.apply_service_api_v2_services_v2_apply_put(
            service_v2_config
        ).result

        current_state = service.goal_state
        self.log.info(
            f"Service {service.id} has been deployed. Service is transitioning towards: {current_state}."
        )
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}'
        )

        if config.canary_percent is not None and len(service.versions) > 1:
            if config.canary_percent == 100:
                self.log.info(
                    "The canary percent has been set to 100%. This will NOT complete the rollout. "
                    "To complete the rollout, submit a rollout command without the --canary-percent flag."
                )
            self._log_canary_version_query(service)

    def _overwrite_runtime_env_in_v2_ray_serve_config(self, config: ServiceConfig):
        """Modifies config in place."""
        ray_serve_config = config.ray_serve_config
        if "applications" in ray_serve_config:
            for ray_serve_app_config in ray_serve_config["applications"]:
                ray_serve_app_config["runtime_env"] = override_runtime_env_config(
                    runtime_env=ray_serve_app_config.get("runtime_env"),
                    anyscale_api_client=self.anyscale_api_client,
                    api_client=self.api_client,
                    workload_type=Workload.SERVICES,
                    compute_config_id=config.compute_config_id,
                    log=self.log,
                )

        else:
            ray_serve_config["runtime_env"] = override_runtime_env_config(
                runtime_env=ray_serve_config.get("runtime_env"),
                anyscale_api_client=self.anyscale_api_client,
                api_client=self.api_client,
                workload_type=Workload.SERVICES,
                compute_config_id=config.compute_config_id,
                log=self.log,
            )

    def _log_canary_version_query(self, service: DecoratedProductionServiceV2APIModel):
        """
        This method logs the canary version's query endpoint.

        Since the url and token from the service object are not initially available,
        we poll until they become available.
        """
        assert len(service.versions) > 1
        # We assume versions are returned in order of creation.
        canary_version = service.versions[1]
        production_job_id = (
            canary_version.production_job_ids[0]
            if canary_version.production_job_ids
            else None
        )
        if production_job_id:
            canary_version_query_printed = False
            for _ in poll(
                interval_secs=_POLL_SERVICE_VERSION_INFO_INTERVAL_IN_SECS,
                timeout_secs=_POLL_SERVICE_VERSION_INFO_TIMEOUT_IN_SECS,
            ):
                try:
                    service_v1 = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
                        production_job_id
                    ).result
                    url = service_v1.url
                    token = service_v1.token
                    service_version_current_state = service_v1.state.current_state
                    if url and token:
                        self.log.info()
                        self.log.info(
                            "You can query the service endpoint using the curl request below:"
                        )
                        self.log.info(f"curl -H 'Authorization: Bearer {token}' {url}")

                        if service_version_current_state in {
                            HaJobStates.PENDING,
                            HaJobStates.AWAITING_CLUSTER_START,
                        }:
                            self.log.info(
                                "Please note the canary version is still pending and may take several minutes to begin successfully returning requests."
                            )
                        canary_version_query_printed = True
                        break
                except Exception:  # noqa: BLE001
                    # We keep re-retrying until the timeout
                    pass
            if not canary_version_query_printed:
                self.log.info(
                    "Please refer to the Service details page if you would like to query the canary version."
                )

    def list(
        self,
        *,
        name: Optional[str] = None,
        service_id: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool = False,
        max_items: int = 10,
    ) -> None:
        self._list_via_service_list_endpoint(
            name=name,
            service_id=service_id,
            project_id=project_id,
            created_by_me=created_by_me,
            max_items=max_items,
        )

    def _list_via_service_list_endpoint(
        self,
        *,
        name: Optional[str] = None,
        service_id: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool = False,
        max_items: int,
    ):
        services = []
        if service_id:
            service_v2_result: DecoratedProductionServiceV2APIModel = (
                self.api_client.get_service_api_v2_services_v2_service_id_get(
                    service_id
                ).result
            )
            services.append(
                [
                    service_v2_result.name,
                    service_v2_result.id,
                    service_v2_result.current_state,
                    # TODO: change to email once https://github.com/anyscale/product/pull/18189 is merged
                    service_v2_result.creator_id,
                ]
            )

        else:
            services_data = self._get_services_by_name(
                name=name,
                project_id=project_id,
                created_by_me=created_by_me,
                max_items=max_items,
            )

            services = [
                [
                    service.name,
                    service.id,
                    service.current_state,
                    service.creator.email,
                ]
                for service in services_data
            ]

        table = tabulate.tabulate(
            services,
            headers=["NAME", "ID", "CURRENT STATE", "CREATOR"],
            tablefmt="plain",
        )
        self.log.info(f'View your Services in the UI at {get_endpoint("/services")}')
        self.log.info(f"Services:\n{table}")

    def archive(self, service_id: str):
        raise click.ClickException(
            f"Archive {service_id} is not currently supported for v2 services."
            "Please contact Anyscale support for more information."
        )

    def rollback(self, service_id: str) -> None:
        service = self.api_client.rollback_service_api_v2_services_v2_service_id_rollback_post(
            service_id
        ).result

        self.log.info(f"Service {service.id} rollback initiated.")
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}'
        )

    def terminate(self, service_id: str) -> None:
        self.api_client.terminate_service_api_v2_services_v2_service_id_terminate_post(
            service_id
        )
        self.log.info(f"Service {service_id} terminate initiated.")
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service_id}")}'
        )



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class LogFileChunk(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cluster_id': 'str',
        'chunk_name': 'str',
        'chunk_url': 'str',
        'size': 'int',
        'file_name': 'str',
        'node_type': 'NodeType',
        'node_ip': 'str',
        'instance_id': 'str',
        'session_id': 'str',
        'ray_pid': 'str',
        'ray_worker_id': 'str',
        'job_id': 'str'
    }

    attribute_map = {
        'cluster_id': 'cluster_id',
        'chunk_name': 'chunk_name',
        'chunk_url': 'chunk_url',
        'size': 'size',
        'file_name': 'file_name',
        'node_type': 'node_type',
        'node_ip': 'node_ip',
        'instance_id': 'instance_id',
        'session_id': 'session_id',
        'ray_pid': 'ray_pid',
        'ray_worker_id': 'ray_worker_id',
        'job_id': 'job_id'
    }

    def __init__(self, cluster_id=None, chunk_name=None, chunk_url=None, size=None, file_name=None, node_type=None, node_ip=None, instance_id=None, session_id=None, ray_pid=None, ray_worker_id=None, job_id=None, local_vars_configuration=None):  # noqa: E501
        """LogFileChunk - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._cluster_id = None
        self._chunk_name = None
        self._chunk_url = None
        self._size = None
        self._file_name = None
        self._node_type = None
        self._node_ip = None
        self._instance_id = None
        self._session_id = None
        self._ray_pid = None
        self._ray_worker_id = None
        self._job_id = None
        self.discriminator = None

        self.cluster_id = cluster_id
        self.chunk_name = chunk_name
        self.chunk_url = chunk_url
        self.size = size
        self.file_name = file_name
        self.node_type = node_type
        self.node_ip = node_ip
        self.instance_id = instance_id
        self.session_id = session_id
        if ray_pid is not None:
            self.ray_pid = ray_pid
        if ray_worker_id is not None:
            self.ray_worker_id = ray_worker_id
        if job_id is not None:
            self.job_id = job_id

    @property
    def cluster_id(self):
        """Gets the cluster_id of this LogFileChunk.  # noqa: E501

        The cluster ID that this file originates from.  # noqa: E501

        :return: The cluster_id of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this LogFileChunk.

        The cluster ID that this file originates from.  # noqa: E501

        :param cluster_id: The cluster_id of this LogFileChunk.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_id`, must not be `None`")  # noqa: E501

        self._cluster_id = cluster_id

    @property
    def chunk_name(self):
        """Gets the chunk_name of this LogFileChunk.  # noqa: E501

        The full object name of a chunk (e.g. .../dashboard_agent.log/chunk-one.log)  # noqa: E501

        :return: The chunk_name of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._chunk_name

    @chunk_name.setter
    def chunk_name(self, chunk_name):
        """Sets the chunk_name of this LogFileChunk.

        The full object name of a chunk (e.g. .../dashboard_agent.log/chunk-one.log)  # noqa: E501

        :param chunk_name: The chunk_name of this LogFileChunk.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and chunk_name is None:  # noqa: E501
            raise ValueError("Invalid value for `chunk_name`, must not be `None`")  # noqa: E501

        self._chunk_name = chunk_name

    @property
    def chunk_url(self):
        """Gets the chunk_url of this LogFileChunk.  # noqa: E501

        A presigned URL to download this chunk.  # noqa: E501

        :return: The chunk_url of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._chunk_url

    @chunk_url.setter
    def chunk_url(self, chunk_url):
        """Sets the chunk_url of this LogFileChunk.

        A presigned URL to download this chunk.  # noqa: E501

        :param chunk_url: The chunk_url of this LogFileChunk.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and chunk_url is None:  # noqa: E501
            raise ValueError("Invalid value for `chunk_url`, must not be `None`")  # noqa: E501

        self._chunk_url = chunk_url

    @property
    def size(self):
        """Gets the size of this LogFileChunk.  # noqa: E501

        The size of this chunk (should never exceed 10MB).  # noqa: E501

        :return: The size of this LogFileChunk.  # noqa: E501
        :rtype: int
        """
        return self._size

    @size.setter
    def size(self, size):
        """Sets the size of this LogFileChunk.

        The size of this chunk (should never exceed 10MB).  # noqa: E501

        :param size: The size of this LogFileChunk.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and size is None:  # noqa: E501
            raise ValueError("Invalid value for `size`, must not be `None`")  # noqa: E501

        self._size = size

    @property
    def file_name(self):
        """Gets the file_name of this LogFileChunk.  # noqa: E501

        The full file name (path) of a file (e.g. .../dashboard_agent.log  # noqa: E501

        :return: The file_name of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._file_name

    @file_name.setter
    def file_name(self, file_name):
        """Sets the file_name of this LogFileChunk.

        The full file name (path) of a file (e.g. .../dashboard_agent.log  # noqa: E501

        :param file_name: The file_name of this LogFileChunk.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and file_name is None:  # noqa: E501
            raise ValueError("Invalid value for `file_name`, must not be `None`")  # noqa: E501

        self._file_name = file_name

    @property
    def node_type(self):
        """Gets the node_type of this LogFileChunk.  # noqa: E501

        The type of node that this file originated from (e.g. head-node, worker-nodes  # noqa: E501

        :return: The node_type of this LogFileChunk.  # noqa: E501
        :rtype: NodeType
        """
        return self._node_type

    @node_type.setter
    def node_type(self, node_type):
        """Sets the node_type of this LogFileChunk.

        The type of node that this file originated from (e.g. head-node, worker-nodes  # noqa: E501

        :param node_type: The node_type of this LogFileChunk.  # noqa: E501
        :type: NodeType
        """
        if self.local_vars_configuration.client_side_validation and node_type is None:  # noqa: E501
            raise ValueError("Invalid value for `node_type`, must not be `None`")  # noqa: E501

        self._node_type = node_type

    @property
    def node_ip(self):
        """Gets the node_ip of this LogFileChunk.  # noqa: E501

        The node IP that this file originated from.  # noqa: E501

        :return: The node_ip of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._node_ip

    @node_ip.setter
    def node_ip(self, node_ip):
        """Sets the node_ip of this LogFileChunk.

        The node IP that this file originated from.  # noqa: E501

        :param node_ip: The node_ip of this LogFileChunk.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and node_ip is None:  # noqa: E501
            raise ValueError("Invalid value for `node_ip`, must not be `None`")  # noqa: E501

        self._node_ip = node_ip

    @property
    def instance_id(self):
        """Gets the instance_id of this LogFileChunk.  # noqa: E501

        The instance ID that this file originated from.  # noqa: E501

        :return: The instance_id of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._instance_id

    @instance_id.setter
    def instance_id(self, instance_id):
        """Sets the instance_id of this LogFileChunk.

        The instance ID that this file originated from.  # noqa: E501

        :param instance_id: The instance_id of this LogFileChunk.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and instance_id is None:  # noqa: E501
            raise ValueError("Invalid value for `instance_id`, must not be `None`")  # noqa: E501

        self._instance_id = instance_id

    @property
    def session_id(self):
        """Gets the session_id of this LogFileChunk.  # noqa: E501

        The session ID that this file originated from.  # noqa: E501

        :return: The session_id of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._session_id

    @session_id.setter
    def session_id(self, session_id):
        """Sets the session_id of this LogFileChunk.

        The session ID that this file originated from.  # noqa: E501

        :param session_id: The session_id of this LogFileChunk.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and session_id is None:  # noqa: E501
            raise ValueError("Invalid value for `session_id`, must not be `None`")  # noqa: E501

        self._session_id = session_id

    @property
    def ray_pid(self):
        """Gets the ray_pid of this LogFileChunk.  # noqa: E501

        The PID that this file originated from, if appropriate.  # noqa: E501

        :return: The ray_pid of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._ray_pid

    @ray_pid.setter
    def ray_pid(self, ray_pid):
        """Sets the ray_pid of this LogFileChunk.

        The PID that this file originated from, if appropriate.  # noqa: E501

        :param ray_pid: The ray_pid of this LogFileChunk.  # noqa: E501
        :type: str
        """

        self._ray_pid = ray_pid

    @property
    def ray_worker_id(self):
        """Gets the ray_worker_id of this LogFileChunk.  # noqa: E501

        The Ray worker ID that this file originated from, if appropriate.  # noqa: E501

        :return: The ray_worker_id of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._ray_worker_id

    @ray_worker_id.setter
    def ray_worker_id(self, ray_worker_id):
        """Sets the ray_worker_id of this LogFileChunk.

        The Ray worker ID that this file originated from, if appropriate.  # noqa: E501

        :param ray_worker_id: The ray_worker_id of this LogFileChunk.  # noqa: E501
        :type: str
        """

        self._ray_worker_id = ray_worker_id

    @property
    def job_id(self):
        """Gets the job_id of this LogFileChunk.  # noqa: E501

        The Ray job ID that this file originated from, if appropriate.  # noqa: E501

        :return: The job_id of this LogFileChunk.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this LogFileChunk.

        The Ray job ID that this file originated from, if appropriate.  # noqa: E501

        :param job_id: The job_id of this LogFileChunk.  # noqa: E501
        :type: str
        """

        self._job_id = job_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LogFileChunk):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LogFileChunk):
            return True

        return self.to_dict() != other.to_dict()



import ipaddress
import json
from pathlib import Path
import re
import sys
from tempfile import NamedTemporaryFile
from typing import Any, Callable, List, Optional
from unittest.mock import MagicMock, Mock, patch

import click
from google.api_core.exceptions import NotFound
from google.cloud.compute_v1.types import Subnetwork
from google.cloud.compute_v1.types.compute import FirewallPolicyRule
from google.iam.v1.policy_pb2 import Binding
from googleapiclient.errors import HttpError
import pytest

from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.models import CreateCloudResourceGCP
from anyscale.client.openapi_client.models.gcp_file_store_config import (
    GCPFileStoreConfig,
)
from anyscale.client.openapi_client.models.gcp_memorystore_instance_config import (
    GCPMemorystoreInstanceConfig,
)
from anyscale.gcp_verification import (
    _check_bucket_region,
    _firewall_rules_from_proto_resp,
    _gcp_subnet_has_enough_capacity,
    _get_proxy_only_subnet_in_vpc,
    _verify_service_account_on_bucket,
    GCPLogger,
    verify_cloud_storage,
    verify_filestore,
    verify_firewall_policy,
    verify_gcp_access_service_account,
    verify_gcp_dataplane_service_account,
    verify_gcp_networking,
    verify_gcp_project,
    verify_memorystore,
)
from anyscale.utils.gcp_utils import GCP_REQUIRED_APIS
from anyscale.utils.network_verification import Direction, FirewallRule, Protocol


_MOCK_GCP_PROJECT_ID = "anyscale-bridge-deadbeef"

_MOCK_GCP_CLOUD_REGION = "us-west1"


def generate_gcp_cloud_mock_resources(**overrides: Any) -> CreateCloudResourceGCP:
    default_resources = {
        "gcp_vpc_id": "vpc_one",
        "gcp_anyscale_iam_service_account_email": "anyscale-admin@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
        "gcp_cluster_node_service_account_email": "cld-xyz-access@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
        "gcp_subnet_ids": ["subnet_one"],
        "gcp_firewall_policy_ids": ["firewall"],
        "gcp_filestore_config": GCPFileStoreConfig(
            instance_name="projects/anyscale-bridge-deadbeef/locations/us-west1/instances/store-one",
            root_dir="mock_root_dir",
            mount_target_ip="mock_ip",
        ),
        "gcp_cloud_storage_bucket_id": "cloud-bucket",
        "memorystore_instance_config": GCPMemorystoreInstanceConfig(
            name="projects/anyscale-bridge-deadbeef/locations/us-west1/instances/memorystore-one",
            endpoint="mock_ip",
        ),
    }
    return CreateCloudResourceGCP(**{**default_resources, **overrides})


@pytest.mark.parametrize(
    ("responses", "expected_result", "num_calls"),
    [
        pytest.param(
            {".*/networks/.*": [("404 Not Found", None)]}, False, 1, id="Network404",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("404 Not Found", None)],
            },
            False,
            2,
            id="SubNet404",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("200 OK", "subnetworks.json")],
            },
            True,
            2,
            id="Success",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("200 OK", "subnetworks_other.json")],
            },
            False,
            2,
            id="WrongNetwrork",
        ),
    ],
)
def test_verify_gcp_networking(
    setup_mock_server, responses, expected_result: bool, num_calls: int
):
    factory, tracker = setup_mock_server
    tracker.reset(responses=responses)
    assert (
        verify_gcp_networking(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )
        == expected_result
    )
    assert len(tracker.seen_requests) == num_calls, tracker.seen_requests


@pytest.mark.parametrize(
    ("cidr", "result", "log"),
    [
        ("10.0.0.0/16", True, ""),
        ("10.0.0.0/19", True, ""),
        ("10.0.0.0/21", True, "We suggest at least"),
        ("10.0.0.0/25", False, "We want at least"),
        pytest.param(
            "",
            False,
            "",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
            id="BadIPAddress",
        ),
    ],
)
def test_gcp_subnet_has_enough_capacity(cidr: str, result: bool, log: str, capsys):
    subnet = Subnetwork(ip_cidr_range=cidr)

    assert _gcp_subnet_has_enough_capacity(subnet, CloudSetupLogger()) == result

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if log:
        assert re.search(log, stderr)


@pytest.mark.parametrize(
    ("responses", "expected"),
    [
        pytest.param(
            {".*/subnetworks.*": [("200 OK", "proxy_only_subnet_present.json")],},
            True,
            id="ProxyOnlySubnetPresent",
        ),
        pytest.param(
            {".*/subnetworks.*": [("200 OK", "proxy_only_subnet_missing.json")],},
            False,
            id="ProxyOnlySubnetMissing",
        ),
        pytest.param(
            {".*/subnetworks.*": [("200 OK", "proxy_only_subnet_no_subnets.json")],},
            False,
            id="ProxyOnlySubnetNoSubnets",
        ),
        pytest.param(
            {".*/subnetworks.*": [("200 OK", "proxy_only_subnet_wrong_vpc.json")],},
            False,
            id="ProxyOnlySubnetWrongVPC",
        ),
    ],
)
def test_get_proxy_only_subnet_in_vpc(setup_mock_server, responses, expected):
    factory, tracker = setup_mock_server
    tracker.reset(responses=responses)
    assert (
        _get_proxy_only_subnet_in_vpc(factory, "myproject", "us-central1", "myvpc")
        is not None
    ) == expected
    assert len(tracker.seen_requests) == 1


@pytest.mark.parametrize(
    "call_factory",
    [
        pytest.param(
            lambda f: f.compute_v1.NetworksClient().get(project="abc", network="cde"),
            id="Compute",
        ),
        pytest.param(
            lambda f: f.resourcemanager_v3.ProjectsClient().get_project(
                name="projects/abc"
            ),
            id="ResourceManager",
        ),
    ],
)
def test_client_factory_cloud_client(setup_mock_server, call_factory: Callable):
    factory, tracker = setup_mock_server
    tracker.reset(responses={".*": [("404 Not Found", None)]})
    with pytest.raises(NotFound):
        call_factory(factory)


def test_client_factory_apis(setup_mock_server):
    factory, tracker = setup_mock_server
    tracker.reset(responses={".*": [("418 I'm a teapot", None)]})

    with pytest.raises(HttpError) as e:
        factory.build("iam", "v1").projects().serviceAccounts().get(
            name="projects/-/serviceAccounts/abc"
        ).execute()
    assert e.value.status_code == 418


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param({".*": [("404 Not Found", None)]}, False, id="NotFound"),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "project_get.json"),
                    ("200 OK", "project_iam_bindings.json"),
                    ("200 OK", "project_apis_enabled.json"),
                ]
            },
            True,
            id="ProjectExists",
        ),
        pytest.param(
            {".*": [("200 OK", "project_get_inactive.json")]},
            False,
            id="ProjectInactive",
        ),
    ],
)
def test_verify_gcp_project(setup_mock_server, responses, expected_result: bool):
    factory, tracker = setup_mock_server
    tracker.reset(responses)
    assert (
        verify_gcp_project(factory, MagicMock(), _MOCK_GCP_PROJECT_ID, _gcp_logger())
        == expected_result
    )


def test_verify_gcp_missing_bindigs(setup_mock_server, capsys):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "project_get.json"),
                ("200 OK", "project_iam_bindings.json"),
                ("200 OK", "project_apis_enabled.json"),
            ]
        }
    )
    assert verify_gcp_project(factory, MagicMock(), _MOCK_GCP_PROJECT_ID, _gcp_logger())
    _, err = capsys.readouterr()
    assert (
        "The Compute Engine Service Agent does not have the standard IAM Role of 'roles/compute.serviceAgent'"
        in err
    )


@pytest.mark.parametrize(
    ("responses", "yes", "expected_result"),
    [
        pytest.param(
            {
                ".*": [
                    ("200 OK", "project_get.json"),
                    ("200 OK", "project_iam_bindings.json"),
                    ("200 OK", "project_apis_enabled.json"),
                ]
            },
            False,
            True,
            id="success",
        ),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "project_get.json"),
                    ("200 OK", "project_iam_bindings.json"),
                    ("200 OK", "project_apis_disabled.json"),
                ]
            },
            False,
            False,
            id="project-disabled",
        ),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "project_get.json"),
                    ("200 OK", "project_iam_bindings.json"),
                    ("200 OK", "project_apis_disabled.json"),
                ]
            },
            True,
            True,
            id="will-enable-apis",
        ),
    ],
)
@pytest.mark.parametrize("has_memorystore", [True, False])
def test_verify_gcp_project_apis(
    setup_mock_server,
    responses,
    has_memorystore: bool,
    yes: bool,
    expected_result: bool,
):
    factory, tracker = setup_mock_server
    tracker.reset(responses)
    mock_resource = MagicMock(
        memorystore_instance_config=Mock() if has_memorystore else None
    )
    mock_enable_project_apis = Mock()
    with patch.multiple(
        "anyscale.gcp_verification",
        enable_project_apis=mock_enable_project_apis,
        click=Mock(confirm=Mock(return_value=False)),
    ):
        assert (
            verify_gcp_project(
                factory, mock_resource, _MOCK_GCP_PROJECT_ID, _gcp_logger(yes)
            )
            == expected_result
        )
        if yes and expected_result:
            mock_enable_project_apis.assert_called_once()
        else:
            mock_enable_project_apis.assert_not_called()
        query_str = "".join(
            [req.get("QUERY_STRING", "") for req in tracker.seen_requests]
        )
        for api in GCP_REQUIRED_APIS:
            assert api in query_str
        assert has_memorystore == ("redis.googleapis.com" in query_str)


@pytest.mark.parametrize(
    ("service_account", "expected_result"),
    [
        pytest.param(
            "anyscale-admin-owner@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            True,
            id="Owner",
        ),
        pytest.param(
            "anyscale-admin-editor@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            True,
            id="Editor",
        ),
        pytest.param(
            "anyscale-admin-editor-not-signer@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            False,
            id="EditorNotSigner",
        ),
        pytest.param(
            "nope@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            False,
            id="NotGranted",
        ),
    ],
)
def test_verify_gcp_access_service_account(
    setup_mock_server, service_account: str, expected_result: bool
):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "access_service_account_permissions.json"),
                ("200 OK", "project_iam_binding_access.json"),
            ]
        }
    )
    mock_resources = generate_gcp_cloud_mock_resources(
        gcp_anyscale_iam_service_account_email=service_account
    )
    if expected_result:
        assert (
            verify_gcp_access_service_account(
                factory, mock_resources, _MOCK_GCP_PROJECT_ID, _gcp_logger(),
            )
            == expected_result
        )
    else:
        with pytest.raises(click.exceptions.Abort):
            verify_gcp_access_service_account(
                factory, mock_resources, _MOCK_GCP_PROJECT_ID, _gcp_logger(yes=False),
            )


def test_verify_gcp_access_service_account_get_iam_policy_error(setup_mock_server):
    factory, tracker = setup_mock_server
    mock_resources = generate_gcp_cloud_mock_resources()
    tracker.reset({".*": [("403 Forbidden", None),]})
    assert (
        verify_gcp_access_service_account(
            factory, mock_resources, _MOCK_GCP_PROJECT_ID, _gcp_logger(yes=False)
        )
        is False
    )


@pytest.mark.parametrize(
    ("use_shared_vpc", "responses", "expected_result", "output"),
    [
        pytest.param(
            False,
            {".*": [("404 Not Found", None), ("404 Not Found", None)]},
            False,
            re.compile(".*"),
            id="Neither Exist",
        ),
        pytest.param(
            False,
            {
                ".*": [
                    ("200 OK", "global_firewall.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            True,
            re.compile(".*"),
            id="GlobalFirewall",
        ),
        pytest.param(
            True,
            {
                ".*": [
                    ("200 OK", "global_firewall.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            True,
            re.compile(".*"),
            id="GlobalFirewall-sharedvpc",
        ),
        pytest.param(
            False,
            {
                ".*": [
                    ("404 Not Found", None),
                    ("200 OK", "regional_firewall.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            True,
            re.compile("Global firweall policy .* not found"),
            id="RegionalFirewall",
        ),
        pytest.param(
            False,
            {".*": [("200 OK", "global_firewall_wrong_vpc.json")],},
            False,
            re.compile(".*is not associated with the VPC.*"),
            id="WrongNetwork",
        ),
    ],
)
def test_verify_firewall_policy(
    setup_mock_server,
    responses,
    use_shared_vpc: bool,
    expected_result: bool,
    output: re.Pattern,
    capsys,
):
    factory, tracker = setup_mock_server
    tracker.reset(responses=responses)

    assert (
        verify_firewall_policy(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            use_shared_vpc,
            False,
            GCPLogger(CloudSetupLogger(), "project", Mock()),
        )
        == expected_result
    )
    _, stderr = capsys.readouterr()
    assert output.search(stderr), f"Missing output in {stderr}"


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param(
            {
                ".*": [
                    ("200 OK", "global_firewall_without_service_rule.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            False,
            id="GlobalFirewall-sharedvpc",
        ),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "global_firewall.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            True,
            id="GlobalFirewall",
        ),
    ],
)
def test_verify_firewall_policy_rule_for_shared_vpc(
    setup_mock_server, responses, expected_result: bool, capsys,
):
    factory, tracker = setup_mock_server
    tracker.reset(responses=responses)

    assert (
        verify_firewall_policy(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            True,
            False,
            GCPLogger(CloudSetupLogger(), "project", Mock()),
        )
        is True
    )
    _, stderr = capsys.readouterr()
    print(stderr)
    assert ("required for Anyscale services" not in stderr) == expected_result


@pytest.mark.parametrize(
    ("responses", "has_subnet", "expected_warning"),
    [
        pytest.param(
            {
                ".*": [
                    ("200 OK", "global_firewall_with_service_rule.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            False,
            "Could not find a subnet with purpose",
            id="has-subnet-no-rule",
        ),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "global_firewall_with_service_rule.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            True,
            "does not allow inbound access from the proxy-only subnet",
            id="has-subnet-no-rule",
        ),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "global_firewall_with_private_service_rule.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            True,
            None,
            id="has-subnet-no-rule",
        ),
    ],
)
def test_verify_firewall_policy_rule_for_private_services_on_shared_vpc(
    setup_mock_server,
    responses,
    expected_warning: Optional[str],
    has_subnet: bool,
    capsys,
):
    factory, tracker = setup_mock_server
    tracker.reset(responses=responses)

    # CIDR is rom the global_firewall_with_private_service_rule.json file
    mock_subnet = Mock(ip_cidr_range="10.100.0.0/20") if has_subnet else None

    with patch.multiple(
        "anyscale.gcp_verification",
        _get_proxy_only_subnet_in_vpc=Mock(return_value=mock_subnet),
    ):
        assert (
            verify_firewall_policy(
                factory,
                generate_gcp_cloud_mock_resources(),
                _MOCK_GCP_PROJECT_ID,
                _MOCK_GCP_CLOUD_REGION,
                True,
                True,
                GCPLogger(CloudSetupLogger(), "project", Mock()),
            )
            is True
        )
    _, stderr = capsys.readouterr()
    if expected_warning:
        assert expected_warning in stderr


@pytest.mark.parametrize(
    ("proto", "rule"),
    [
        (FirewallPolicyRule(action="deny"), []),
        (
            FirewallPolicyRule(
                direction="EGRESS",
                action="allow",
                match={"src_ip_ranges": ["10.10.10.0/24"]},
            ),
            [
                FirewallRule(
                    direction=Direction.EGRESS,
                    protocol=Protocol.all,
                    network=ipaddress.ip_network("10.10.10.0/24"),
                    ports=None,
                )
            ],
        ),
        (
            FirewallPolicyRule(
                direction="EGRESS",
                action="allow",
                match={
                    "src_ip_ranges": ["10.10.10.0/24"],
                    "layer4_configs": [{"ip_protocol": "tcp"}],
                },
            ),
            [
                FirewallRule(
                    direction=Direction.EGRESS,
                    protocol=Protocol.tcp,
                    network=ipaddress.ip_network("10.10.10.0/24"),
                    ports=None,
                )
            ],
        ),
    ],
)
def test_firewall_rules_from_proto_resp(
    proto: FirewallPolicyRule, rule: List[FirewallRule]
):
    assert _firewall_rules_from_proto_resp([proto]) == rule
    assert _firewall_rules_from_proto_resp([proto, proto]) == rule * 2


def test_verify_gcp_access_service_account_does_not_exist(setup_mock_server):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [("404 Not Found", None)]})
    assert not verify_gcp_access_service_account(
        factory,
        generate_gcp_cloud_mock_resources(),
        _MOCK_GCP_PROJECT_ID,
        _gcp_logger(),
    )


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param({".*": [("404 Not Found", None)]}, False, id="DoesNotExist"),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "dataplane_service_account.json"),
                    ("200 OK", "project_iam_binding_access.json"),
                ]
            },
            True,
            id="DoesNotExist",
        ),
    ],
)
def test_verify_gcp_dataplane_service_account(
    setup_mock_server, responses, expected_result: bool, capsys
):
    factory, tracker = setup_mock_server
    tracker.reset(responses)
    assert (
        verify_gcp_dataplane_service_account(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _gcp_logger(),
        )
        == expected_result
    )

    _, err = capsys.readouterr()
    if expected_result:
        # Check if there is a warning about missing artifact Registry warnings if we pass verification.
        assert re.search("roles/artifactregistry.reader`", err) and re.search(
            "safe to ignore if you are not using", err
        )


def test_verify_gcp_dataplane_service_account_wrong_project(setup_mock_server, capsys):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "dataplane_service_account.json"),
                ("200 OK", "project_iam_binding_access.json"),
            ]
        }
    )
    assert verify_gcp_dataplane_service_account(
        factory, generate_gcp_cloud_mock_resources(), "something-else", _gcp_logger(),
    )

    _, stderr = capsys.readouterr()
    assert (
        "constraints/iam.disableCrossProjectServiceAccountUsage` is not enforced"
        in stderr
    )


@pytest.mark.parametrize(
    ("response", "result"),
    [
        pytest.param(("200 OK", "regional_filestore.json"), True, id="regional"),
        pytest.param(("200 OK", "zonal_filestore.json"), True, id="zonal"),
        pytest.param(
            ("200 OK", "regional_filestore_wrong_vpc.json"), False, id="wrong_vpc"
        ),
        pytest.param(
            ("200 OK", "regional_filestore_vpc_private_access.json"),
            True,
            id="vpc_private_access",
        ),
        pytest.param(("404 Not Found", None), False, id="not_found"),
    ],
)
def test_verify_filestore(setup_mock_server, response, result: bool):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [response]})
    assert (
        verify_filestore(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )
        == result
    )


@pytest.mark.parametrize(
    ("response", "has_location_warning"),
    [
        pytest.param(
            ("200 OK", "regional_filestore_west_region.json"),
            False,
            id="EnterpriseWestZone",
        ),
        pytest.param(
            ("200 OK", "regional_filestore.json"), True, id="EnterpriseCentralZone",
        ),
        pytest.param(
            ("200 OK", "zonal_filestore.json"), True, id="NonEnterpriseCentralZone",
        ),
    ],
)
def test_verify_filestore_warn(
    setup_mock_server, capsys, response, has_location_warning: bool,
):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [response]})
    assert verify_filestore(
        factory,
        generate_gcp_cloud_mock_resources(),
        _MOCK_GCP_CLOUD_REGION,
        _gcp_logger(),
    )
    _, err = capsys.readouterr()
    assert ("cross-region" in err) == has_location_warning


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param([("404 Not Found", None)], False, id="NoBucket"),
        pytest.param(
            [
                ("200 OK", "storage_bucket.json"),
                ("200 OK", "storage_bucket_policy.json"),
            ],
            True,
            id="Success",
        ),
    ],
)
def test_verify_cloud_storage(setup_mock_server, responses, expected_result: bool):
    factory, tracker = setup_mock_server
    tracker.reset({".*": responses})
    assert (
        verify_cloud_storage(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )
        == expected_result
    )


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param([("404 Not Found", None)], False, id="NoInstance"),
        pytest.param(
            [("200 OK", "regional_memorystore.json"),], False, id="VPCMismatch",
        ),
        pytest.param(
            [("200 OK", "regional_memorystore.json"),], False, id="tierMismatch",
        ),
        pytest.param(
            [("200 OK", "regional_memorystore_standard_tier.json"),],
            False,
            id="misconfiguredMaxmemory",
        ),
        pytest.param(
            [("200 OK", "regional_memorystore_read_replica_disabled.json"),],
            False,
            id="readReplicaDisabled",
        ),
        pytest.param(
            [("200 OK", "regional_memorystore_tls_enabled.json"),],
            False,
            id="tlsDisabled",
        ),
        pytest.param(
            [("200 OK", "regional_memorystore_happy.json"),], True, id="happy",
        ),
    ],
)
def test_verify_memorystore(
    setup_mock_server, responses, expected_result: bool, request
):
    factory, tracker = setup_mock_server
    cloud_resources = generate_gcp_cloud_mock_resources()

    if request.node.callspec.id != "vpcMismatch":
        # This is the vpc id used in frontend/cli/tests/gcp_responses/regional_filestore.json
        cloud_resources.gcp_vpc_id = "anyscale-vpn-cloud"

    tracker.reset({".*": responses})
    assert (
        verify_memorystore(factory, cloud_resources, _gcp_logger(), strict=True,)
        == expected_result
    )


@patch.multiple(
    "anyscale.gcp_verification", ANYSCALE_HOST="https://console.anyscale.com"
)
@pytest.mark.parametrize(
    ("update_fn", "expected_warn"),
    [
        pytest.param(
            lambda x: x["iamConfiguration"]["uniformBucketLevelAccess"].update(
                {"enabled": False}
            ),
            "does not have Uniform Bucket Access enabled,",
            id="UniformBucketAccess",
        ),
        pytest.param(
            lambda x: x["iamConfiguration"].update(
                {"publicAccessPrevention": "unspecified"}
            ),
            "public access prevention",
            id="PublicAccessPrevention",
        ),
        pytest.param(
            lambda x: x.update({"cors": []}), "create the correct CORS rule", id="CORS"
        ),
    ],
)
def test_verify_cloud_storage_configuration_warnings(
    setup_mock_server, update_fn: Callable, expected_warn: str, capsys
):
    factory, tracker = setup_mock_server

    with NamedTemporaryFile(mode="w") as bad_bucket:
        responses = [
            ("200 OK", "storage_bucket.json"),
            ("200 OK", "storage_bucket_policy.json"),
            ("200 OK", bad_bucket.name),
            ("200 OK", "storage_bucket_policy.json"),
        ]
        with open(
            Path(__file__).parent.joinpath("gcp_responses", "storage_bucket.json")
        ) as f:
            existing = json.load(f)
            update_fn(existing)
            bad_bucket.write(json.dumps(existing))
            bad_bucket.flush()

        tracker.reset({".*": responses})
        assert verify_cloud_storage(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )

        # First call should not have an error
        _, err = capsys.readouterr()
        assert expected_warn not in err

        # Second call should have the error
        assert verify_cloud_storage(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )
        _, err = capsys.readouterr()
        assert expected_warn in err


def test_verify_cloud_storage_improper_service_account_configuration(
    setup_mock_server, capsys
):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "storage_bucket.json"),
                ("200 OK", "storage_bucket_policy.json"),
            ]
            * 2
        }
    )

    common_error_text = (
        "other_email@project.com requires the following permissions on Bucket "
    )
    mock_resource = generate_gcp_cloud_mock_resources(
        gcp_anyscale_iam_service_account_email="other_email@project.com"
    )
    assert verify_cloud_storage(
        factory,
        mock_resource,
        _MOCK_GCP_PROJECT_ID,
        _MOCK_GCP_CLOUD_REGION,
        _gcp_logger(),
    )
    _, err = capsys.readouterr()
    assert err.count(common_error_text) == 1

    mock_resource = generate_gcp_cloud_mock_resources(
        gcp_cluster_node_service_account_email="other_email@project.com"
    )
    assert verify_cloud_storage(
        factory,
        mock_resource,
        _MOCK_GCP_PROJECT_ID,
        _MOCK_GCP_CLOUD_REGION,
        _gcp_logger(),
    )

    _, err = capsys.readouterr()
    assert err.count(common_error_text) == 1


@pytest.mark.parametrize(
    ("bindings", "expected_result"),
    [
        pytest.param(
            [Binding(role="roles/storage.admin", members=["sa@email.com"])],
            True,
            id="Admin",
        ),
        pytest.param(
            [
                Binding(role="roles/storage.objectAdmin", members=["sa@email.com"]),
                Binding(
                    role="roles/storage.legacyBucketReader", members=["sa@email.com"]
                ),
            ],
            True,
            id="Legacy+Admin",
        ),
        pytest.param(
            [Binding(role="roles/storage.legacyBucketOwner", members=["sa@email.com"])],
            False,
            id="MissingObjectRead",
        ),
        pytest.param(
            [
                Binding(
                    role="roles/storage.legacyBucketOwner", members=["sa@email.com"]
                ),
                Binding(role="roles/storage.objectAdmin", members=["sa_two@email.com"]),
            ],
            False,
            id="NecessaryPermissionOnCorrect",
        ),
    ],
)
def test_verify_service_account_on_bucket(bindings: List[Binding], expected_result):
    assert (
        _verify_service_account_on_bucket("sa@email.com", bindings) == expected_result
    )
    assert not _verify_service_account_on_bucket("dne@email.com", bindings)


@pytest.mark.parametrize(
    ("region", "file", "result"),
    [
        ("us-west1", "storage_bucket.json", True),
        ("us-east1", "storage_bucket.json", True),
        ("europe-central2", "storage_bucket.json", False),
        ("europe-central2", "storage_dual_region.json", True),
        ("europe-central1", "storage_dual_region.json", False),
        ("southamerica-west1", "storage_single_region.json", True),
        ("southamerica-east1", "storage_single_region.json", False),
    ],
)
def test_check_bucket_region(setup_mock_server, region: str, file: str, result: bool):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [("200 OK", file)]})
    assert (
        _check_bucket_region(
            factory.storage.Client("project").get_bucket("bucket"), region
        )[0]
        == result
    )


def _gcp_logger(yes: bool = True) -> GCPLogger:
    return GCPLogger(CloudSetupLogger(), "project_abc", Mock(), yes=yes)
